from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import uuid

from .auid import AUID
from .model import datadefs
from . import core
from .utils import register_class

def short_name(name):
    for s in ('DataDef_', 'ContainerDef_'):
        name = name.replace(s, "")
    return name

def lookup_def(dictionary, name, instance_type, key):
    if isinstance(name, AUID):
        return dictionary[key].get(name, None)

    if isinstance(name, uuid.UUID):
        return dictionary[key].get(AUID(name), None)

    if isinstance(name, instance_type):
        return name

    name = short_name(name).lower()

    for key, value in dictionary[key].items():
        value_name = value.short_name or ''
        if name == value_name.lower():
            return value

    raise Exception("No Definition: %s" % str(name))

@register_class
class DefinitionObject(core.AAFObject):
    class_id = AUID("0d010101-0101-1a00-060e-2b3402060101")
    __slots__ = ()
    def __init__(self, auid=None, name=None, description=None):
        super(DefinitionObject, self).__init__()
        self.name = name
        self.description = description
        if auid:
            self.auid = AUID(str(auid))

    @property
    def name(self):
        return self['Name'].value

    @name.setter
    def name(self, value):
        self['Name'].value = value

    @property
    def short_name(self):
        name = self.name
        if name:
            return short_name(name)

    @property
    def description(self):
        return self['Description'].value

    @description.setter
    def description(self, value):
        self['Description'].value =  value

    @property
    def auid(self):
        return self['Identification'].value

    @auid.setter
    def auid(self, value):
        self['Identification'].value = value

    @property
    def uuid(self):
        auid = self.auid
        if auid is not None:
            return auid.uuid

    @uuid.setter
    def uuid(self, value):
        self.auid = AUID(value)

    def __repr__(self):
        s = "%s.%s" % (self.__class__.__module__,
                       self.__class__.__name__)
        name = self.name
        if name:
            s += " %s" % name

        return '<%s %s at 0x%x>' % (s, str(self.auid), id(self))

    @property
    def unique_key(self):
        return self.auid

@register_class
class DataDef(DefinitionObject):
    class_id = AUID("0d010101-0101-1b00-060e-2b3402060101")
    __slots__ = ()

@register_class
class OperationDef(DefinitionObject):
    class_id = AUID("0d010101-0101-1c00-060e-2b3402060101")
    __slots__ = ()

    @property
    def datadef(self):
        return self['DataDefinition'].value

    @datadef.setter
    def datadef(self, value):
        self['DataDefinition'].value = value

    @property
    def media_kind(self):
        datadef = self.datadef
        if datadef:
            return datadef.short_name

    @media_kind.setter
    def media_kind(self, value):
        self.datadef = self.root.dictionary.lookup_datadef(value)

    @property
    def parameters(self):
        return self['ParametersDefined']

    @property
    def number_inputs(self):
        return self['NumberInputs'].value

    @number_inputs.setter
    def number_inputs(self, value):
        self['NumberInputs'].value = value


@register_class
class ParameterDef(DefinitionObject):
    class_id = AUID("0d010101-0101-1d00-060e-2b3402060101")
    __slots__ = ()

    def __init__(self, auid=None, name=None, description=None, typedef=None):
        super(ParameterDef, self).__init__(auid, name, description)
        if typedef:
            self.typedef = self.root.dictionary.lookup_typedef(typedef)

    @property
    def typedef(self):
        return self['Type'].value

    @typedef.setter
    def typedef(self, value):
        self['Type'].value = value

@register_class
class PluginDef(DefinitionObject):
    class_id = AUID("0d010101-0101-1e00-060e-2b3402060101")
    __slots__ = ()

@register_class
class CodecDef(DefinitionObject):
    class_id = AUID("0d010101-0101-1f00-060e-2b3402060101")
    __slots__ = ()
    def __init__(self, dictionary, auid=None, name=None, description=None, classdef=None, datadef_names=None):
        super(CodecDef, self).__init__(auid, name, description)
        if classdef:
            self['FileDescriptorClass'].value = self.root.metadict.lookup_classdef(classdef)

        for d in datadef_names or []:
            self['DataDefinitions'].append(dictionary.lookup_datadef(d))

@register_class
class ContainerDef(DefinitionObject):
    class_id = AUID("0d010101-0101-2000-060e-2b3402060101")
    __slots__ = ()

@register_class
class InterpolationDef(DefinitionObject):
    class_id = AUID("0d010101-0101-2100-060e-2b3402060101")
    __slots__ = ()

@register_class
class TaggedValueDef(DefinitionObject):
    class_id = AUID("0d010101-0101-4c00-060e-2b3402060101")
    __slots__ = ()

@register_class
class Dictionary(core.AAFObject):
    class_id = AUID("0d010101-0101-2200-060e-2b3402060101")
    __slots__ = ()
    def __init__(self):
        super(Dictionary, self).__init__()

        for key, args in datadefs.DataDefs.items():
            d = self.root.create.DataDef(key, *args)
            self['DataDefinitions'].append(d)

        for key, args in datadefs.ContainerDefs.items():
            d = self.root.create.ContainerDef(key, *args)
            self['ContainerDefinitions'].append(d)

    def setup_defaults(self):

        for key, args in datadefs.CodecDefs.items():
            if len(args) > 2:
                d = self.root.create.CodecDef(self, key, *args)
                self['CodecDefinitions'].append(d)

    def register_def(self, defobject):

        if isinstance(defobject, DataDef):
            self['DataDefinitions'].append(defobject)
        elif isinstance(defobject, ContainerDef):
            self['ContainerDefinitions'].append(defobject)
        elif isinstance(defobject, CodecDef):
            self['CodecDefinitions'].append(defobject)
        elif isinstance(defobject, ParameterDef):
            self['ParameterDefinitions'].append(defobject)
        elif isinstance(defobject, OperationDef):
            self['OperationDefinitions'].append(defobject)
        elif isinstance(defobject, InterpolationDef):
            self['InterpolationDefinitions'].append(defobject)
        elif isinstance(defobject, TaggedValueDef):
            self['TaggedValueDefinitions'].append(defobject)
        else:
            raise ValueError("unknown definitions type: %s" % str(type(defobject)))

    def lookup_typedef(self, name):
        return self.root.metadict.lookup_typedef(name)

    def lookup_datadef(self, name):
        return lookup_def(self, name, DataDef, 'DataDefinitions')

    def lookup_containerdef(self, name):
        return lookup_def(self, name, ContainerDef, 'ContainerDefinitions')

    def lookup_codecdef(self, name):
        return lookup_def(self, name, CodecDef, 'CodecDefinitions')

    def lookup_parameterdef(self, name):
        return lookup_def(self, name, ParameterDef, 'ParameterDefinitions')

    def lookup_operationdef(self, name):
        return lookup_def(self, name, OperationDef, 'OperationDefinitions')

    def lookup_interperlationdef(self, name):
        return lookup_def(self, name, InterpolationDef, 'InterpolationDefinitions')

    def lookup_taggedvaluedef(self, name):
        return lookup_def(self, name, TaggedValueDef, 'TaggedValueDefinitions')

    def update(self, other):
        if not isinstance(other, Dictionary):
            raise ValueError("other must be a Dictionary object")

        for def_type in ('DataDefinitions',
                         'ContainerDefinitions',
                         'CodecDefinitions',
                         'ParameterDefinitions',
                         'OperationDefinitions',
                         'InterpolationDefinitions',
                         'TaggedValueDefinitions'):

            for other_def in other[def_type].values():
                if other_def.auid not in self[def_type]:
                    self[def_type].append(other_def.copy(root=self.root))

                elif def_type == 'OperationDefinitions':
                    # make sure operation all the same parameters
                    opdef = self.lookup_operationdef(other_def.auid)
                    for p in other_def.parameters:
                        param = self.lookup_parameterdef(p.auid)
                        if param.unique_key not in opdef.parameters:
                            opdef.append(param)



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
from aaf2.cfb import CompoundFileBinary
import os
import io

import common
import shutil

base = os.path.join(os.path.dirname(os.path.abspath(__file__)))
test_dir = os.path.join(base, 'results')
if not os.path.exists(test_dir):
    os.makedirs(test_dir)

test_files = os.path.join(base, 'test_files')

def small_data(count):
    for i in range(count):
        data = b"some data %d" % i
        yield  data * 10

def large_data(count):
    for i in range(count):
        data = b"some larger data %d" % i
        yield  data * 10000

import unittest
class StreamTests(unittest.TestCase):

    def write_and_test(self, filename, data_list, chunksize=61):
        path = os.path.join(test_dir, filename)

        with io.open(path, 'wb+', buffering=io.DEFAULT_BUFFER_SIZE) as f:
            ss = CompoundFileBinary(f, 'wb+')

            for i, data in enumerate(data_list):
                s = ss.open("/test_stream%d" % i, 'w')
                write_data = data
                while len(write_data):
                    s.write(write_data[:chunksize])
                    write_data = write_data[chunksize:]

                s.seek(0)

                result = s.read()

                assert result == data

                s.close()

            for i, data in enumerate(data_list):
                s = ss.open("/test_stream%d" % i, 'r')
                s.seek(0)
                result = s.read()

                assert result == data

            ss.close()


        with io.open(path, 'rb', buffering=io.DEFAULT_BUFFER_SIZE) as f:

            ss = CompoundFileBinary(f)

            for i, data in enumerate(data_list):
                s = ss.open("/test_stream%d" % i, 'r')

                assert s.dir.byte_size, len(data)

                result = s.read()

                assert result == data

    def write_and_ovewrite(self, filename, data_list1, data_list2, chunksize=61):

        path = os.path.join(test_dir, filename)

        with io.open(path, 'wb+', buffering=io.DEFAULT_BUFFER_SIZE) as f:

            ss = CompoundFileBinary(f, 'wb+')

            for i, data in enumerate(data_list1):
                s = ss.open("/test_stream%d" % i, 'w')
                while len(data):
                    s.write(data[:chunksize])
                    data = data[chunksize:]
                s.close()

            for i, data in enumerate(data_list2):
                s = ss.open("/test_stream%d" % i, 'w')
                while len(data):
                    s.write(data[:chunksize])
                    data = data[chunksize:]
                s.close()

            ss.close()

        with io.open(path, 'rb',  buffering=io.DEFAULT_BUFFER_SIZE) as f:

            ss = CompoundFileBinary(f)

            for i, data in enumerate(data_list2):
                s = ss.open("/test_stream%d" % i, 'r')
                assert s.dir.byte_size, len(data)

                result = s.read()
                assert result == data

    def test_mini_stream(self):
        self.write_and_test("mini_stream.aaf", [b"small data\n" * 10])

    def test_multi_mini_stream(self):
        self.write_and_test("mulit_mini_stream.aaf", [b"small1 data\n" * 10, b"small2 data\n" * 10])

    def test_large_stream(self):
        self.write_and_test("large_stream.aaf", [b"large data\n" * 5000])

    def test_large_mini(self):
        self.write_and_test("mini_and_large_stream.aaf", [b"large data\n" * 5000,
                                                          b"small data\n" * 10])
    def test_mini_large(self):
        self.write_and_test("mini_and_large_stream.aaf", [b"small data\n" * 10,
                                                          b"large data\n" * 5000
                                                     ])

    def test_lots_of_small(self):
        self.write_and_test("lots_of_small_streams.aaf", list(small_data(300)))

    def test_lots_of_large(self):
        self.write_and_test("lots_of_large_streams.aaf", list(large_data(100)), chunksize=1024)

    def test_64(self):
        self.write_and_test("64.aaf", list(small_data(4)), chunksize=64)

    def test_4096(self):
        self.write_and_test("4096.aaf", list(large_data(5)), chunksize=4096)


    def test_mix_bag(self):

        s = list(small_data(50))
        l = list(large_data(50))
        d = []
        for x,y in zip(s, l):
            d.append(x)
            d.append(y)

        self.write_and_test("mixed_bag_streams.aaf", d, chunksize=1024)


    def test_overwrite(self):

        data1 = [b"small data\n" * 20,
                 b"large data\n" * 5000]

        data2 = [b"overwrite data\n" * 10,
                 b"overwrite data\n" * 10000]

        self.write_and_ovewrite("mini_and_large_stream.aaf", data1, data2)

    def test_seek(self):
        path = os.path.join(test_dir, "seek_test.aaf")

        with open(path, 'wb+') as f:
            ss = CompoundFileBinary(f, 'wb+')

            s = ss.open("/seektest", 'w')
            s.seek(100)
            data= b"end of seek"
            s.write(data)
            assert s.tell() == 100 + len(data)
            s.close()

            s = ss.open("/seek_large", 'w')
            seek_size = 1000000
            s.seek(seek_size)
            data= b"end of seek"
            s.write(data)
            assert s.dir.byte_size == seek_size  + len(data)
            s.close()
            ss.close()

    def test_move(self):
        path = os.path.join(test_dir, "move_test.aaf")

        src = "/path/to/item"
        dst = "/dest/path/moved_item"

        stream_data = b'move stream data'

        steam_list = [
        "/path/to/item/child1/sub/sub/stream",
        "/path/to/item/stream",
        "/path/to/item2/stream"
        ]

        dir_list = [
        "/path/to/item/child1/sub/sub",
        "/path/to/item/child2",
        "/path/to/item/child3/more_stuff",
        "/path/to/item1/child1",
        "/path/to/item2/child2/sub/sub",
        "/path/to/item2/child2",
        "/dest/path/",
        ]

        result_dirs = [
        "/dest",
        "/dest/path",
        "/dest/path/moved_item",
        "/dest/path/moved_item/child1",
        "/dest/path/moved_item/child1/sub",
        "/dest/path/moved_item/child1/sub/sub",
        "/dest/path/moved_item/child3",
        "/path",
        "/path/to",
        "/path/to/item1",
        "/path/to/item2",
        "/path/to/item2/child2",
        "/path/to/item2/child2/sub",
        ]

        result_streams = [
            "/dest/path/moved_item/child1/sub/sub/stream",
            "/dest/path/moved_item/stream",
            '/dest/path/moved_item/child3/stream'
        ]

        with open(path, 'wb+') as f:
            cfb = CompoundFileBinary(f, 'wb+')
            for p in dir_list:
                cfb.makedirs(p)

            for p in steam_list:
                s = cfb.open(p, 'w')
                s.write(stream_data)

            cfb.move(src, dst)
            with self.assertRaises(ValueError):
                cfb.move('/path/to/item2/stream', '/dest/path/moved_item/child3')

            cfb.move('/path/to/item2/stream', '/dest/path/moved_item/child3/')

            with self.assertRaises(ValueError):
                cfb.move('/path/that/doesnt/exist', '/dest/path/moved_item/')


            with self.assertRaises(ValueError):
                cfb.move('/path/to/item2/child2', '/path/that/doesnt/exist')


            for p in result_dirs:
                assert cfb.exists(p)
                entry = cfb.find(p)
                assert entry.isdir()

            for p in result_streams:
                assert cfb.exists(p)
                entry = cfb.find(p)
                assert entry.isfile()
                assert entry.open('r').read() == stream_data

            cfb.close()

        with open(path, 'rb') as f:
            cfb = CompoundFileBinary(f, 'rb')

            for root, storage_items, streams in cfb.walk():
                pass
                # print(root.path(), streams)

            for p in result_dirs:
                assert cfb.exists(p)
                entry = cfb.find(p)
                assert entry.isdir()

            for p in result_streams:
                assert cfb.exists(p)
                entry = cfb.find(p)
                assert entry.isfile()

                assert entry.open('r').read() == stream_data

    def test_move2(self):
        src_file = common.test_file_01()
        test_file = os.path.join(test_dir, "move2_test.aaf")
        shutil.copy(src_file, test_file)

        dir_list = []
        stream_list = []
        move_root = "/Header-2/Content-3b03"

        with open(test_file, 'rb+') as f:
            cfb = CompoundFileBinary(f, 'rb+')
            for root, storage_items, streams in cfb.walk(move_root):

                for item in storage_items:
                    dir_list.append(item.path())
                for item in streams:
                    stream_list.append(item.path())

            cfb.makedir('/tmp')
            cfb.move(move_root, '/tmp/')

            for item in dir_list:
                assert not cfb.exists(item)
                new_path = item.replace(move_root, '/tmp/Content-3b03')
                assert cfb.exists(new_path)
                assert cfb.find(new_path).isdir()

            for item in stream_list:
                assert not cfb.exists(item)
                new_path = item.replace(move_root, '/tmp/Content-3b03')
                assert cfb.exists(new_path)
                assert cfb.find(new_path).isfile()

            cfb.close()

        with open(test_file, 'rb') as f:
            cfb = CompoundFileBinary(f, 'rb')
            for item in dir_list:
                assert not cfb.exists(item)
                new_path = item.replace(move_root, '/tmp/Content-3b03')
                assert cfb.exists(new_path)
                assert cfb.find(new_path).isdir()

            for item in stream_list:
                assert not cfb.exists(item)
                new_path = item.replace(move_root, '/tmp/Content-3b03')
                assert cfb.exists(new_path)
                assert cfb.find(new_path).isfile()


    def test_rmtree(self):
        src_file = common.test_file_01()
        test_file = os.path.join(test_dir, "rmtree_test.aaf")
        shutil.copy(src_file, test_file)

        remove_list = []
        remove_root = "/Header-2/Content-3b03"

        property_streams = []

        with open(test_file, 'rb+') as f:
            cfb = CompoundFileBinary(f, 'rb+')
            # for root, storage_items, streams in cfb.walk():
            #     for item in streams:
            #         print("!!", item.path())

            for root, storage_items, streams in cfb.walk(remove_root):

                for item in storage_items:
                    remove_list.append(item.path())
                for item in streams:
                    remove_list.append(item.path())

            cfb.rmtree('/Header-2/Content-3b03')

            for item in remove_list:
                assert not cfb.exists(item)

            for root, storage_items, streams in cfb.walk():
                for item in streams:
                    if item.name == 'properties':
                        property_streams.append(item.path())

            for item in property_streams:
                cfb.remove(item)

            for item in property_streams:
                assert not cfb.exists(item)

            cfb.close()

        with open(test_file, 'rb') as f:
            cfb = CompoundFileBinary(f, 'rb')
            for item in remove_list:
                assert not cfb.exists(item)

            for item in property_streams:
                assert not cfb.exists(item)

    def test_trunc_zero(self):
        test_file = os.path.join(test_dir, "trunc_zero_test.aaf")
        with open(test_file, 'wb+') as f:
            cfb = CompoundFileBinary(f, 'wb+')
            s = cfb.open("/zero_trunc", 'w')
            s.write(b'some data')
            s.seek(0)
            s.truncate()
            cfb.close()

        with open(test_file, 'rb') as f:
            cfb = CompoundFileBinary(f, 'rb')
            s = cfb.open("/zero_trunc", 'r')
            assert s.dir.byte_size == 0
            assert s.dir.sector_id is None


    def test_trunc_shrink(self):
        test_file = os.path.join(test_dir, "trunc_shrink_test.aaf")
        with open(test_file, 'wb+') as f:
            cfb = CompoundFileBinary(f, 'w+')
            s = cfb.open("/grow_trunc", 'w')
            data = b"F" * 5000
            s.write(data)
            assert s.is_mini_stream() == False
            s.truncate(256)
            assert s.is_mini_stream() == True
            cfb.close()
        with open(test_file, 'rb') as f:
            cfb = CompoundFileBinary(f, 'rb')
            s = cfb.open("/grow_trunc", 'r')
            data = s.read()
            assert data == b"F" * 256


if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)


    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import unittest
import common
import os
import aaf2

class EssenceTests(unittest.TestCase):

    def test_create(self):
        new_file = os.path.join(common.sandbox(), 'create_essence.aaf')
        test_data = b"Essence Data!!"
        mob_id =  aaf2.mobid.MobID()
        mob_id.int = 1

        with aaf2.open(new_file, 'w') as f:
            mob = f.create.SourceMob()
            mob.descriptor = f.create.TapeDescriptor()
            f.content.mobs.append(mob)

            e = f.create.EssenceData()
            e.mob = mob

            f.content.essencedata.append(e)
            stream = e.open("w")
            stream.write(test_data)

            e.mob = mob

            mob.mob_id = mob_id
            e.mob_id = mob_id


        with aaf2.open(new_file, 'r') as f:
            e = next(f.content.essencedata.values())
            assert e.mob_id == mob_id
            stream = e.open("r")
            assert stream.read() == test_data
            mob = e.mob
            assert mob

    def test_write_dettached(self):
        new_file = os.path.join(common.sandbox(), 'essencedata_dettached_write.aaf')
        test_data = b"Essence Data!!"
        mob_id =  aaf2.mobid.MobID()
        mob_id.int = 1

        with aaf2.open(new_file, 'w') as f:
            e = f.create.EssenceData()
            e.mob_id = mob_id
            stream = e.open("w")
            stream.write(test_data)

            f.content.essencedata.append(e)

            stream = e.open('r')
            assert stream.read() == test_data

        with aaf2.open(new_file, 'r') as f:
            e = next(f.content.essencedata.values())
            stream = e.open("r")
            assert stream.read() == test_data

            assert not f.cfb.exists("/tmp")

    def test_write_dettached_clean(self):
        new_file = os.path.join(common.sandbox(), 'essencedata_dettached_clean.aaf')
        test_data = b"Essence Data!!"
        mob_id =  aaf2.mobid.MobID()
        mob_id.int = 1

        with aaf2.open(new_file, 'w') as f:
            e = f.create.EssenceData()
            e.mob_id = mob_id
            stream = e.open("w")
            stream.write(test_data)

        with aaf2.open(new_file, 'r') as f:
            assert not f.cfb.exists("/tmp")

    def test_read_exception(self):

        new_file = os.path.join(common.sandbox(), 'read_exception.aaf')
        test_data = b"Essence Data!!"
        mob_id =  aaf2.mobid.MobID()
        mob_id.int = 1

        with aaf2.open(new_file, 'w') as f:
            e = f.create.EssenceData()
            e.mob_id = mob_id

            with self.assertRaises(aaf2.exceptions.AAFPropertyError):
                stream = e.open('r')

            f.content.essencedata.append(e)

        with aaf2.open(new_file, 'r') as f:
            e = next(f.content.essencedata.values())
            with self.assertRaises(aaf2.exceptions.AAFPropertyError):
                stream = e.open('r')



if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import os
from aaf2.file import AAFFile
from aaf2.cfb import CompoundFileBinary
from aaf2.mobid import MobID
from aaf2 import exceptions
import aaf2

import unittest

import common
import datetime

class CreatAAFTests(unittest.TestCase):

    def dump(self, path):
        with AAFFile(path, 'r') as f:
            for root, storage, streams in f.cfb.walk():
                print(root.path())

    def test_empty(self):

        result_file = common.get_test_file("empty.aaf")
        with AAFFile(result_file, 'w') as f:
            pass


        with AAFFile(result_file, 'r') as f:
            assert f.metadict
            assert f.content
            assert f.header['ObjectModelVersion'].value == 1
            assert f.header['Version'].value == {u'major': 1, u'minor': 2}


    def test_empty_512(self):

        result_file = common.get_test_file("empty_512.aaf")
        with AAFFile(result_file, 'w', sector_size=512) as f:
            pass


        with AAFFile(result_file, 'r') as f:
            assert f.metadict
            assert f.content
            assert f.header['ObjectModelVersion'].value == 1
            assert f.header['Version'].value == {u'major': 1, u'minor': 2}

        # self.dump(result_file)

    def test_mobs(self):

        result_file = common.get_test_file('mobs.aaf')

        mobs  = {}
        now = datetime.datetime.now()
        count = 100
        with AAFFile(result_file, 'w') as f:

            for i in range(count):
                mob_id = MobID.new()
                m = f.create.MasterMob()
                m.name = "TestMob%d" %i
                m.mob_id = mob_id
                m['LastModified'].value = now
                m['CreationTime'].value = now
                m['Slots'].value = []

                f.content.mobs.append(m)

                mobs[mob_id] = m.name

        with AAFFile(result_file, 'r') as f:
            # file_mobs = f.content['Mobs'].value

            for mob in f.content.mobs:
                assert mob.mob_id in mobs
                assert mob.name == mobs[mob.mob_id]

            assert len(list(f.content.mobs)) == count


    def test_abstract(self):

        with AAFFile() as f:
            # try and create some abstract classes
            with self.assertRaises(ValueError):
                f.create.Segment()

            with self.assertRaises(ValueError):
                f.create.Component()

            with self.assertRaises(ValueError):
                f.create.PhysicalDescriptor()

    def test_delete_propetry(self):
        result_file = common.get_test_file('delete.aaf')
        with aaf2.open(result_file, 'w') as f:
            m = f.create.MasterMob()
            m['AppCode'].value = 14
            m.usage = 'Usage_LowerLevel'
            f.content.mobs.append(m)

        with aaf2.open(result_file, 'rw') as f:
            mob = list(f.content.mobs)[0]
            assert 'AppCode' in mob
            assert 'UsageCode' in mob

            del mob['AppCode']
            del mob['UsageCode']

            assert 'AppCode' not in mob
            assert 'UsageCode' not in mob

            with self.assertRaises(KeyError):
                del mob['AppCode']



        with aaf2.open(result_file, 'r') as f:
            mob = list(f.content.mobs)[0]
            assert 'AppCode' not in mob
            assert 'UsageCode' not in mob

            with self.assertRaises(KeyError):
                del mob['AppCode']

    def test_bad_timestamp(self):
        result_file = common.get_test_file('bad_timestamp.aaf')
        bad_timestamp = {u'date': {u'month': 0, u'day': 0, u'year': 0},
                         u'time': datetime.time(0, 0)}

        # NOTE: in versions 1.1.0 and 1.0.2 of the AAF SDK it is
        # possible to have datetimes python cannot decode
        # SourceForge Bug ID #1191791
        # https://sourceforge.net/p/aaf/bugs/90/

        with aaf2.open(result_file, 'w') as f:
            f.header['LastModified'].value = bad_timestamp

        with aaf2.open(result_file, 'r') as f:
            assert f.header['LastModified'].value == bad_timestamp

if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import unittest
import aaf2
import common

class TestCreateSequence(unittest.TestCase):

    def test_create_sequence(self):

        result_file = common.get_test_file('create_sequence.aaf')
        mob_count = 0
        components = 0

        with aaf2.open(result_file, "w")  as f:

            video_rate = " 30000/1001"

            comp_mob = f.create.CompositionMob()
            sequence = f.create.Sequence(media_kind="picture")

            timeline_slot = comp_mob.create_timeline_slot(video_rate)
            timeline_slot.segment= sequence

            f.content.mobs.append(comp_mob)

            length = 60 * 30
            filler_len = 100
            timecode_fps = 30
            mob_count += 1

            test_path = "some_path.mov"

            for i in range(10):

                # Make the Tape MOB
                tape_mob = f.create.SourceMob()
                tape_name = "tape_name"
                tape_slot, tape_timecode_slot = tape_mob.create_tape_slots(tape_name, video_rate, timecode_fps)
                tape_slot.segment.length = length

                f.content.mobs.append(tape_mob)
                mob_count += 1
                # Make a FileMob
                file_mob = f.create.SourceMob()

                # Make a locator
                loc = f.create.NetworkLocator()
                loc['URLString'].value = test_path

                file_description = f.create.CDCIDescriptor()
                file_description.locator.append(loc)

                file_description['ComponentWidth'].value = 8
                file_description['HorizontalSubsampling'].value = 4
                file_description['ImageAspectRatio'].value = '16/9'
                file_description['StoredWidth'].value = 1920
                file_description['StoredHeight'].value = 1080
                file_description['FrameLayout'].value = 'FullFrame'
                file_description['VideoLineMap'].value = [42, 0]
                file_description['SampleRate'].value = video_rate
                file_description['Length'].value = 10

                file_mob.descriptor = file_description

                clip = tape_mob.create_source_clip(slot_id=1, length=length)
                slot = file_mob.create_picture_slot(video_rate)
                slot.segment.components.append(clip)


                f.content.mobs.append(file_mob)
                mob_count += 1

                # Make the Master MOB
                master_mob = f.create.MasterMob()
                master_mob.name = "Master Mob %i" % i

                master_mob.comments['Test'] = 'Value'
                master_mob.comments.append(f.create.TaggedValue("Test2", 42))

                assert master_mob.comments['Test'] == "Value"
                assert master_mob.comments['Test2'] == 42

                clip = file_mob.create_source_clip(slot_id=1)
                assert clip.length == length
                slot = master_mob.create_picture_slot(video_rate)
                slot.segment.components.append(clip)

                f.content.mobs.append(master_mob)
                mob_count += 1

                # Create a SourceClip
                clip = master_mob.create_source_clip(slot_id=1)
                assert clip.length == length
                sequence.components.append(clip)
                components += 1

                # Create a filler
                comp_fill = f.create.Filler("picture", filler_len)
                sequence.components.append(comp_fill)
                components += 1

        with aaf2.open(result_file, "r")  as f:
            assert len(f.content.mobs) == mob_count
            comp = next(f.content.compositionmobs())
            slot = comp.slot_at(1)
            assert len(slot.segment.components) == components




if __name__ == "__main__":
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import os
import aaf2

import unittest

import common
import datetime

class TaggedValueTests(unittest.TestCase):


    def test_tagged_value(self):
        result_file = common.get_test_file('tagged_value.aaf')

        with aaf2.open(result_file, 'w') as f:

            t1 = f.create.TaggedValue("name1", 3221)

            typedef = f.dictionary.lookup_typedef("aafInt32")
            assert t1.value_typedef == typedef

            t2 =  f.create.TaggedValue("name2", 45, 'aafInt64')

            typedef = f.dictionary.lookup_typedef("aafInt64")
            assert t2.value_typedef == typedef

            t3 = f.create.TaggedValue()

            t3.name = "name3"
            t3.encode_value("24000/10001", 'Rational')

            typedef = f.dictionary.lookup_typedef("Rational")
            assert t3.value_typedef == typedef



if __name__ == "__main__":
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import sys
from io import BytesIO
import struct
import array

from .utils import (
    write_u8,
    write_u16le,
    safe_print,
    )
from .exceptions import AAFPropertyError, AAFAttachError
from . import properties
from .properties import property_formats
from .auid import AUID

P_HEADER_STRUCT = struct.Struct(str('<BBH'))

OPERATIONGROUP_PARAMETERS_AUID = AUID("06010104-060a-0000-060e-2b3401010102")

sentinel = object()


class AAFObject(object):
    __slots__ = ('class_id', 'root', 'dir', 'property_entries', '__weakref__' )

    def __new__(cls, *args, **kwargs):
        self = super(AAFObject, cls).__new__(cls)
        self.root = None
        self.dir = None
        self.property_entries = {}
        return self

    def __init__(self, *args, **kwargs):
        pass

    @property
    def classdef(self):
        return self.root.metadict.classdefs_by_auid.get(self.class_id, None)

    @property
    def name(self):
        classdef = self.classdef
        if classdef:
            return classdef.class_name
        return self.__class__.__name__

    @property
    def unique_property(self):
        classdef = self.classdef
        if classdef:
            pid = self.classdef.unique_key_pid
            if pid:
                p = self.property_entries.get(pid, None)
                if p:
                    return p

        raise AAFPropertyError("Object has no unique property")

    @property
    def unique_key(self):
        return self.unique_property.value

    def read_properties(self):
        stream = self.dir.get('properties')
        if stream is None:
            return

        s = stream.open()

        # read the whole stream
        with BytesIO(s.read()) as f:
            (byte_order, version, entry_count) = P_HEADER_STRUCT.unpack(f.read(4))

            if byte_order != 0x4c:
                raise NotImplementedError("be byteorder")

            props = array.array(str('H'))
            if hasattr(props, 'frombytes'):
                props.frombytes(f.read(6 * entry_count))
            else:
                props.fromstring(f.read(6 * entry_count))

            if sys.byteorder == 'big':
                props.byteswap()

            for i in range(entry_count):
                index = i * 3
                pid = props[index + 0]
                format = props[index + 1]
                byte_size = props[index + 2]

                data = f.read(byte_size)
                p = property_formats[format](self, pid, format, version)
                p.data = data
                self.property_entries[pid] = p

            for p in self.property_entries.values():
                p.decode()
                if isinstance(p, (properties.StrongRefSetProperty,
                                  properties.StrongRefVectorProperty,
                                  properties.WeakRefArrayProperty)):
                    p.read_index()

    def validate(self):
        missing = []

        for propertydef in self.classdef.all_propertydefs():

            if propertydef.optional:
                continue

            if propertydef.pid in self.property_entries:
                continue

            # stored on dir entry I guess...
            if propertydef.property_name in ('ObjClass',):
                continue

            missing.append(propertydef)

        if missing:
            m = []
            for p in missing:
                m.append("%s %s" % (p.property_name, str(p.typedef)))
            message = "%s missing the following required properties:\n    %s" % (str(self), "\n    ".join(m))
            raise AAFPropertyError(message)

    def write_properties(self, validate=True):
        if validate:
            self.validate()

        s = self.dir.touch("properties").open(mode='rw')

        with BytesIO() as f:
            byte_order = 0x4c
            entry_count = len(self.property_entries)
            version = properties.PROPERTY_VERSION

            write_u8(f, byte_order)
            write_u8(f, version)
            write_u16le(f, entry_count)

            for p in self.property_entries.values():
                assert p.data is not None
                write_u16le(f, p.pid)
                write_u16le(f, p.format)
                write_u16le(f, len(p.data))

            # write the data
            for p in self.property_entries.values():
                f.write(p.data)

            s.write(f.getvalue())
            s.truncate()
            # write index's
            for p in self.property_entries.values():
                if isinstance(p, (properties.StrongRefSetProperty,
                                  properties.StrongRefVectorProperty,
                                  properties.WeakRefArrayProperty)):
                    p.write_index()

    def detach(self, delete=False):
        items = []
        for item, streams in self.walk_references(topdown=True):
            # store reference not sure if necessary
            items.append(item)

            for pid, p in item.property_entries.items():
                if isinstance(p, (properties.StrongRefProperty,
                                  properties.StrongRefVectorProperty,
                                  properties.StrongRefSetProperty,
                                  properties.StreamProperty)):
                    # converts all internal weakrefs to strongrefs
                    p.detach()

            if item.dir:
                # remove child from object manager
                self.root.manager.pop(item.dir.path(), None)

                # NOTE: do we need to delete directory entry?
                # new objects will overwrite whats there anyway
                # and nothing will point to the old object
                if delete:
                    self.root.cfb.rmtree(item.dir)

                item.dir = None

        # remove self from object manager
        if self.dir:
            self.root.manager.pop(self.dir.path(), None)
            self.dir = None

    def attach(self, dir_entry):
        if self.dir:
            raise AAFAttachError("cannot attached obj to %s already attached to %s" % (dir_entry.path(), self.dir.path()))

        self.dir = dir_entry
        self.dir.class_id = self.class_id

        # add self to modified
        self.root.manager.add_modified(self)

        for pid, p in self.property_entries.items():
            if isinstance(p, (properties.StrongRefProperty,
                              properties.StrongRefVectorProperty,
                              properties.StrongRefSetProperty,
                              properties.StreamProperty)):

                p.attach()

    def walk_references(self, topdown=False):
        refs = []
        streams = []
        for p in self.properties():
            if isinstance(p, properties.StrongRefProperty):
                obj = p.value
                if obj:
                    refs.append(obj)

            elif isinstance(p, properties.StrongRefVectorProperty):
                for v in p.value:
                    refs.append(v)

            elif isinstance(p, properties.StrongRefSetProperty):

                for obj in p.value:
                    refs.append(obj)

            elif isinstance(p, properties.StreamProperty):
                streams.append(p)

        if not topdown:
            yield self, streams

        for obj in refs:
            for item, item_streams in obj.walk_references(topdown):
                yield item, item_streams

        if topdown:
            yield self, streams

    def copy(self, new_dir=None, root=None, classdef_cache=None):
        new_obj = self.__class__.__new__(self.__class__)
        classdef_cache = classdef_cache or set()
        root = root or self.root
        new_obj.root = root

        # if copying to new root make sure classdef is defined
        # to prevent excessive calls a cache is used skip already
        # processed classdefs
        if root is not self.root:
            classdef = self.classdef
            if classdef.auid not in classdef_cache:
                root.metadict.register_external_classdef(self.classdef)
                classdef_cache.add(classdef.auid)

        if type(new_obj) is AAFObject:
            new_obj.class_id = self.class_id
        new_obj.dir = new_dir
        if new_dir:
            assert new_dir.storage is root.cfb
            new_obj.dir.class_id = self.class_id

        for pid, p in self.property_entries.items():
            # the pid in the new root could be different
            if pid >= 0x8000 and root is not self.root:
                pdef = new_obj.classdef.lookup_propertydef(p.propertydef.auid)
                assert pdef
                pid = pdef.pid

            new_obj.property_entries[pid] = p.copy(new_obj, pid, classdef_cache)

        return new_obj

    def properties(self):

        for pid, p in self.property_entries.items():
            yield p

    def allkeys(self):
        keys = []
        classdef = self.classdef
        if not classdef:
            return keys
        for propertydef in classdef.all_propertydefs():
            keys.append(propertydef.property_name)
        return keys

    def keys(self):
        keys = []
        for pid, p in self.property_entries.items():
            keys.append(p.name)
        return keys

    def get(self, key, default=None, allkeys=True):
        for item in self.properties():
            if item.name == key:
                return item
        if not allkeys:
            return default

        classdef = self.classdef
        for propertydef in classdef.all_propertydefs():
            if propertydef.property_name == key:
                fmt = propertydef.store_format

                # Workaround, for OperationDef Parameters
                # AAF SDK uses a StrongRefSetProperty
                # Spec says its suppose to be a StrongRefVectorProperty
                if propertydef.auid == OPERATIONGROUP_PARAMETERS_AUID:
                    fmt = properties.SF_STRONG_OBJECT_REFERENCE_SET

                p = property_formats[fmt](self, propertydef.pid, fmt)
                return p

        return default

    def getvalue(self, key, default=None):
        if key not in self.keys():
            return default

        p = self.get(key, sentinel)
        if p is sentinel:
            return default
        return p.value

    def __getitem__(self, key):
        result = self.get(key, default=sentinel)
        if result is sentinel:
            raise KeyError(key)
        return result

    def __delitem__(self, key):
        result = self.get(key, default=sentinel, allkeys=False)
        if result is sentinel:
            raise KeyError(key)

        del self.property_entries[result.pid]
        if self.dir:
            self.root.manager.add_modified(self)

    def __contains__(self, key):
        result = self.get(key, default=sentinel, allkeys=False)
        if result is sentinel:
            return False
        return True

    def __repr__(self):
        s = "%s.%s" % (self.__class__.__module__,
                       self.__class__.__name__)
        name = self.name
        if name:
            s += ' %s' % name

        return '<%s at 0x%x>' % (s, id(self))

    def dump(self, space=""):
        indent = "  "

        for p in self.properties():

            if isinstance(p, properties.StrongRefProperty):
                safe_print(space, p.name, p.typedef)

                p.value.dump(space + indent)

            if isinstance(p, properties.StrongRefVectorProperty):
                safe_print(space, p.name, p.typedef)
                for obj in p.value:
                    safe_print(space + indent, obj)
                    obj.dump(space + indent*2)
                continue

            if isinstance(p, properties.StrongRefSetProperty):
                safe_print(space, p.name, p.typedef)
                for key, obj in p.items():
                    safe_print(space + indent, obj)
                    obj.dump(space + indent*2)
                continue

            safe_print(space, p.name, p.typedef, p.value)



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import os
import unittest
import aaf2
from aaf2.file import AAFFile
from aaf2 import properties

import common
import shutil

class AAFTests(unittest.TestCase):


    def test_walk_all(self):
        test_file = os.path.join(common.test_files_dir(),"test_file_01.aaf")
        with AAFFile(test_file) as f:
            common.walk_aaf(f.root)

    def test_save_as(self):

        new_file = os.path.join(common.sandbox(), 'save_r+.aaf')
        test_file = common.test_file_01()
        shutil.copy(test_file, new_file)

        with aaf2.open(new_file, 'r+') as f:
            f.save()


        # should contents compare!
        with aaf2.open(new_file, 'r') as f:
            common.walk_aaf(f.root)

    def test_exit_with_exception_no_save(self):
        new_file = os.path.join(common.sandbox(), 'test_exit_with_exception_no_save.aaf')
        test_file = common.test_file_01()
        shutil.copy(test_file, new_file)

        test_mob_id = None
        with self.assertRaises(ValueError):
            with aaf2.open(new_file, 'r+') as f:
                mob = f.create.MasterMob("TestExitMob")
                test_mob_id = mob.mob_id
                f.content.mobs.append(mob)

                raise ValueError('asd')

        self.assertTrue(f.f.closed)
        self.assertFalse(f.is_open)
        self.assertFalse(f.cfb.is_open)

        with open(test_file, 'rb') as testFile:
            with open(new_file, 'rb') as newFile:
                self.assertNotEqual(testFile.read(), newFile.read())

        with aaf2.open(new_file, 'r') as f:
            # Make sure that the file can still be opened with aaf2.open
            # and that the save method wasn't run.
            self.assertIsNone(f.content.mobs.get(test_mob_id, None))

    def test_exit_with_exception_with_save(self):

        new_file = os.path.join(common.sandbox(), 'test_exit_with_exception_with_save.aaf')
        test_file = common.test_file_01()
        shutil.copy(test_file, new_file)

        test_mob_id = None

        with self.assertRaises(ValueError):
            with aaf2.open(new_file, 'r+') as f:
                mob = f.create.MasterMob("TestExitMob")
                test_mob_id = mob.mob_id
                f.content.mobs.append(mob)

                f.save()

                raise ValueError('asd')

        self.assertTrue(f.f.closed)
        self.assertFalse(f.is_open)
        self.assertFalse(f.cfb.is_open)

        with open(test_file, 'rb') as testFile:
            with open(new_file, 'rb') as newFile:
                self.assertNotEqual(testFile.read(), newFile.read())

        with aaf2.open(new_file, 'r') as f:
            # Make sure that the file can still be opened with aaf2.open
            # and that the save method wasn't run.
            for mob in f.content.mobs:
                print(mob)
            self.assertIsNotNone(f.content.mobs.get(test_mob_id, None))

    def test_exit_with_internal_exception(self):
        """This scenario is a bad situation... explosion
        Exercise the scenario where no exception occurs inside with with block, but
        an exception occurs during the exit phase of the context manager.
        """

        new_file = os.path.join(common.sandbox(), 'test_exit_with_internal_exception.aaf')
        test_file = common.test_file_01()
        shutil.copy(test_file, new_file)

        with self.assertRaises(RuntimeError):
            with aaf2.open(new_file, 'r+') as f:
                def mock(*args, **kwargs):
                    raise RuntimeError('asd')

                # This will result in a corrupted file
                f.cfb.close = mock

                # mobs = f.content.mobs
                mob = f.create.MasterMob("TestExitMob")
                f.content.mobs.append(mob)

                f.save()

        self.assertFalse(f.f.closed)
        self.assertTrue(f.is_open)
        self.assertTrue(f.cfb.is_open)

        with open(test_file, 'rb') as testFile:
            with open(new_file, 'rb') as newFile:
                self.assertNotEqual(testFile.read(), newFile.read())

        with self.assertRaises(IndexError):
            # new_file is now corrupted.
            with aaf2.open(new_file, 'r') as f:
                pass

    def test_exit_with_internal_and_external_exception(self):
        # Exception occurs in with block and also in the except clause
        # while closing file descriptors.

        new_file = os.path.join(common.sandbox(), 'test_exit_with_internal_and_external_exception.aaf')
        test_file = common.test_file_01()
        shutil.copy(test_file, new_file)

        # Notice how the exception propagated is not a ValueError.
        with self.assertRaises(RuntimeError):
            with aaf2.open(new_file, 'r+') as f:
                def mock(*args, **kwargs):
                    raise RuntimeError('asd')

                # This will result in a corrupted file
                f.cfb.close = mock

                # mobs = f.content.mobs
                mob = f.create.MasterMob("TestExitMob")
                f.content.mobs.append(mob)

                f.save()

                raise ValueError('asd')

        # And also the file is still closed even if the cfb failed to be closed.
        self.assertTrue(f.f.closed)
        self.assertFalse(f.is_open)
        self.assertTrue(f.cfb.is_open)

        with open(test_file, 'rb') as testFile:
            with open(new_file, 'rb') as newFile:
                self.assertNotEqual(testFile.read(), newFile.read())

        with self.assertRaises(IndexError):
            # new_file is now corrupted.
            with aaf2.open(new_file, 'r') as f:
                pass

    def test_raise_on_close_in_except(self):
        # Test that AAFFile.f.close exceptions are propagated to
        # the user's code.
        with self.assertRaises(RuntimeError):
            with aaf2.open() as fd:
                def mock(*args, **kwargs):
                    raise RuntimeError('asd')

                originalClose = fd.f.close
                fd.f.close = mock
                raise ValueError('asd')

        originalClose()

    def test_save_after_close(self):
        aaf_file = aaf2.open()
        aaf_file.close()
        with self.assertRaises(IOError):
            aaf_file.save()


if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from io import BytesIO
from . import types
from .model import classdefs
from .model import typedefs as base_typedefs
from . import properties
from .auid import AUID

from . import core
from .utils import (register_class, read_u16le, decode_utf16le,
                    encode_utf16le, encode_u16le, str2auid,
                    AAFClaseID_dict, AAFClassName_dict)

PID_NAME      = 0x0006
PID_AUID      = 0x0005
PID_TYPE      = 0x000B
PID_OPTIONAL  = 0x000C
PID_PID       = 0x000D
PID_UNIQUE    = 0x000E

PID_PARENT     = 0x0008
PID_PROPERTIES = 0x0009
PID_CONCRETE   = 0x000A

sentinel = object()


@register_class
class PropertyDef(core.AAFObject):
    class_id = AUID("0d010101-0202-0000-060e-2b3402060101")
    __slots__ = ('_typedef_id', '_auid', '_property_name')

    def __new__(cls, root=None, name=None, auid=None, pid=None, typedef=None, optional=None, unique=None):
        self = super(PropertyDef, cls).__new__(cls)
        self.root = root
        self._typedef_id = None
        self._auid = None
        self._property_name = name
        if root:
            properties.add_string_property(self, PID_NAME, name)
            properties.add_bool_property(self, PID_OPTIONAL, optional)
            properties.add_bool_property(self, PID_UNIQUE, unique)
            properties.add_u16le_property(self, PID_PID, pid)
            properties.add_auid_property(self, PID_AUID, auid)
            properties.add_auid_property(self, PID_TYPE, typedef)

        return self

    @property
    def property_name(self):
        if self._property_name:
            return self._property_name

        p = self.property_entries.get(PID_NAME)
        self._property_name = decode_utf16le(p.data)
        return self._property_name

    @property_name.setter
    def property_name(self, value):
       self.property_entries[PID_NAME].data = encode_utf16le(value)
       self._property_name = value

    @property
    def unique_key(self):
        return self.auid

    @property
    def unique(self):
        if PID_UNIQUE in self.property_entries:
            return self.property_entries[PID_UNIQUE].data == b"\x01"
        return False

    @property
    def pid(self):
        with BytesIO(self.property_entries[PID_PID].data) as f:
            return read_u16le(f)

    @pid.setter
    def pid(self, value):
        self.property_entries[PID_PID].data = encode_u16le(value)

    @property
    def auid(self):
        if self._auid:
            return self._auid
        p = self.property_entries.get(PID_AUID)
        self._auid = AUID(bytes_le=p.data)
        return self._auid

    @property
    def uuid(self):
        return self.auid.uuid

    @property
    def optional(self):
        return self.property_entries[PID_OPTIONAL].data == b"\x01"

    @property
    def typedef_id(self):
        if self._typedef_id:
            return self._typedef_id

        p = self.property_entries.get(PID_TYPE)
        self._typedef_id = AUID(bytes_le=p.data)
        return self._typedef_id

    @property
    def typedef(self):
        type_id = self.typedef_id
        if type_id:
            return self.root.metadict.lookup_typedef(type_id)

    @property
    def store_format(self):
        return self.typedef.store_format

    def __repr__(self):
        return "<%s PropertyDef>" % self.property_name


@register_class
class ClassDef(core.AAFObject):
    class_id = AUID("0d010101-0201-0000-060e-2b3402060101")
    __slots__ = ('propertydef_by_pid')

    def __new__(cls, root=None, name=None, class_auid=None, parent_auid=None, concrete=None):
        self = super(ClassDef, cls).__new__(cls)
        self.root = root
        self.propertydef_by_pid = {}
        if root:
            properties.add_string_property(self, PID_NAME, name)
            properties.add_auid_property(self, PID_AUID, class_auid)
            properties.add_bool_property(self, PID_CONCRETE, concrete)

            # reference self auid if no parent
            properties.add_classdef_weakref_property(self, PID_PARENT, parent_auid or class_auid)
            properties.add_strongref_set_property(self, PID_PROPERTIES, "Properties", PID_AUID)

        return self

    def __eq__(self, other):
        self.auid == other.auid

    @property
    def auid(self):
        data = self.property_entries[PID_AUID].data
        if data is not None:
            return AUID(bytes_le=self.property_entries[PID_AUID].data)

    @property
    def uuid(self):
        return self.auid.uuid

    @property
    def concrete(self):
        return self.property_entries[PID_CONCRETE].data == b"\x01"

    @property
    def class_name(self):
        data = self.property_entries[PID_NAME].data
        if data is not None:
            return decode_utf16le(data)

    @property
    def unique_key(self):
        return self.auid

    @property
    def unique_key_pid(self):
        for p in self.all_propertydefs():
            if p.unique:
                return p.pid

        # Parameter work around
        # Uses the DefinitionObject Identification PID
        if self.isinstance(self.root.metadict.lookup_classdef(AUID("0d010101-0101-3c00-060e-2b3402060101"))):
            return 0x1B01

    @property
    def unique_key_size(self):
        mob_classdef = self.root.metadict.lookup_classdef(AUID("0d010101-0101-3400-060e-2b3402060101"))
        essencedata_classdef = self.root.metadict.lookup_classdef(AUID("0d010101-0101-2300-060e-2b3402060101"))
        if self.isinstance(mob_classdef) or self.isinstance(essencedata_classdef):
            return 32
        return 16

    def isinstance(self, other):

        if self.auid == other.auid:
            return True

        for classdef in other.relatives():
            if classdef.auid == self.auid:
                return True
        return False

    @property
    def name(self):
        return "ClassDefinition"

    @property
    def classdef(self):
        return self.root.metadict.classdefs_by_name['ClassDefinition']

    @property
    def parent_id(self):
        if PID_PARENT in self.property_entries:
            return self.property_entries[PID_PARENT].ref

    @property
    def parent(self):
        parent_id = self.parent_id

        if parent_id == self.auid:
            return None

        return self.root.metadict.classdefs_by_auid.get(parent_id, None)

    @property
    def propertydefs(self):
        if PID_PROPERTIES in self.property_entries:
            return self.property_entries[PID_PROPERTIES].values()
        return []

    def lookup_propertydef(self, property_auid):
        for classdef in self.relatives():
            p = classdef.property_entries.get(PID_PROPERTIES, {}).get(property_auid, None)
            if p:
                return p

    def register_propertydef(self, name, property_auid, pid, typedef, optional, unique=False):

        # check if PropertyDef is already defined
        property_auid = str2auid(property_auid)
        p = self.property_entries.get(PID_PROPERTIES, {}).get(property_auid, None)
        if p:
            return p

        typedef = str2auid(typedef)
        if isinstance(typedef, AUID):
            typedef_auid = typedef
        else:
            typedef = self.root.metadict.lookup_typedef(typedef)
            typedef_auid = typedef.auid

        # None signifies Dynamically allocated Local Tags
        # I think this is similar to MXF where pids > 0x8000 < 0xFFFF
        # are extensions pids
        if pid is None:
            pid = self.root.metadict.next_free_pid()
            # assert False

        p = PropertyDef(self.root, name, property_auid, pid, typedef_auid, optional, unique)
        # # this is done low level to avoid recursion errors

        if PID_PROPERTIES not in self.property_entries:
            properties.add_strongref_set_property(self, PID_PROPERTIES, "Properties", PID_AUID)

        properties.add2set(self, PID_PROPERTIES, p.auid, p)
        self.propertydef_by_pid[pid] = p
        return p

    def relatives(self):
        root = self
        while root:
            yield root
            root = root.parent

    def all_propertydefs(self):
        for classdef in self.relatives():

            if not classdef:
                break
                # raise Exception(self.class_name)

            for p in classdef.propertydefs:
                yield p

    def get_propertydef_from_pid(self, pid, default=None):

        for classdef in self.relatives():
            if not classdef:
                break
            p = classdef.propertydef_by_pid.get(pid, sentinel)

            if p is sentinel:
                 continue
            return p

        return default

    def __repr__(self):
        return "<%s %s>" % (self.class_name, "ClassDef")


root_classes = {
'Root' : ('b3b398a5-1c90-11d4-8053-080036210804', None, True, {
    "Header"              : ('0d010301-0102-0100-060e-2b3401010102', 0x0002, '05022800-0000-0000-060E-2B3401040101', False, False),
    "MetaDictionary"      : ('0d010301-0101-0100-060e-2b3401010102', 0x0001, '05022700-0000-0000-060E-2B3401040101', False, False),
})
}

root_types = {
"HeaderStrongRefence"             : ('05022800-0000-0000-060E-2B3401040101', "0d010101-0101-2f00-060e-2b3402060101"),
"MetaDictionaryStrongReference"   : ('05022700-0000-0000-060E-2B3401040101', "0d010101-0225-0000-060e-2b3402060101"),
}

PID_CLASSDEFS = 0x0003
PID_TYPEDEFS  = 0x0004


@register_class
class MetaDictionary(core.AAFObject):
    class_id = AUID("0d010101-0225-0000-060e-2b3402060101")
    def __init__(self, root):
        super(MetaDictionary, self).__init__(root)

        self.root = root
        root.metadict = self

        properties.add_strongref_set_property(self, PID_CLASSDEFS, "ClassDefinitions", PID_AUID)
        properties.add_strongref_set_property(self, PID_TYPEDEFS, "TypeDefinitions", PID_AUID)

        self.classdefs_by_name = {}
        self.classdefs_by_auid = {}
        self.typedefs_by_name = {}
        self.typedefs_by_auid = {}

        self.typedefs_classes = {}

        self.local_pids = set()
        self.next_pid = 0xFFFF

        for name, args in root_classes.items():
            self.register_classdef(name, *args)

        for name, args in classdefs.classdefs.items():
            self.register_classdef(name, *args)

        for alias, name in classdefs.aliases.items():
            self.classdefs_by_name[alias]= self.classdefs_by_name[name]

        # setup typedefs
        self.register_typedef_model({'strongrefs': root_types})
        self.register_typedef_model(base_typedefs.__dict__)

    def register_typedef_model(self, typedef_model):

        for cat, classobj in types.categories.items():
            for name, args in typedef_model.get(cat, {}).items():

                if not isinstance(args, (tuple,list)):
                    args = [args]

                if cat == "extenums":
                    if name in self.typedefs_by_name:
                        t = self.typedefs_by_name[name]
                        for element_value, element_name in args[1].items():
                            t.register_element(element_name, AUID(element_value))
                    else:
                        t = types.TypeDefExtEnum(self.root, name, *args)
                        properties.add2set(self, PID_TYPEDEFS, t.auid, t)
                elif cat == "enums":
                    if name in self.typedefs_by_name:
                        t = self.typedefs_by_name[name]
                        for element_value, element_name in args[2].items():
                            t.register_element(element_name, element_value)
                    else:
                        t = types.TypeDefEnum(self.root, name, *args)
                        properties.add2set(self, PID_TYPEDEFS, t.auid, t)
                else:

                    t = classobj(self.root, name, *args)
                    properties.add2set(self, PID_TYPEDEFS, t.auid, t)

                self.typedefs_by_name[name] = t
                self.typedefs_by_auid[t.auid] = t
                self.typedefs_classes[t.auid] = t.__class__

    def register_extensions(self):
        from .model.ext import classdefs as ext_classdefs
        from .model.ext import typedefs as ext_typedefs

        for name, args in ext_classdefs.classdefs.items():
            self.register_classdef(name, *args)

        for alias, name in ext_classdefs.aliases.items():
            self.classdefs_by_name[alias]= self.classdefs_by_name[name]

        self.register_typedef_model(ext_typedefs.__dict__)

    def register_classdef(self, name, class_auid, parent, concrete, propertydefs=None):
        if not isinstance(class_auid, AUID):
            class_auid = AUID(class_auid)

        parent = str2auid(parent)

        if parent is None:
            parent_auid = None
        elif isinstance(parent, AUID):
            parent_auid = parent
        else:
            parent_classdef = self.lookup_classdef(parent)
            parent_auid = parent_classdef.auid

        if name in self.classdefs_by_name:
            c = self.classdefs_by_name[name]
        elif class_auid in self.classdefs_by_auid:
            c = self.classdefs_by_name[class_auid]
        else:
            c = ClassDef(self.root, name, class_auid, parent_auid, concrete)

        if propertydefs:
            for prop_name, prop_args in propertydefs.items():
                property_auid = prop_args[0]
                pid = prop_args[1]

                if pid is None:
                    pid = self.next_free_pid()
                elif pid and pid >= 0x8000:
                    assert pid not in self.local_pids
                    self.local_pids.add(pid)

                c.register_propertydef(prop_name, property_auid, pid, *prop_args[2:])

        self.classdefs_by_name[name]   = c
        self.classdefs_by_auid[c.auid] = c

        # Root is not include in the data model for some reason
        if c.class_name != "Root":
            properties.add2set(self, PID_CLASSDEFS, c.auid, c)
        return c

    def register_external_typedef(self, ext_typedef):
        if not isinstance(ext_typedef, types.TypeDef):
            return None

        typedef = self.lookup_typedef(ext_typedef.auid)

        # create typedef if missing
        if not typedef:
            if isinstance(ext_typedef, (types.TypeDefStrongRef,
                                        types.TypeDefWeakRef)):
                # could this cause RecursionError?
                classdef = ext_typedef.ref_classdef
                self.register_external_classdef(classdef)

            elif isinstance(ext_typedef, (types.TypeDefFixedArray,
                                          types.TypeDefVarArray,
                                          types.TypeDefSet,
                                          types.TypeDefString,
                                          types.TypeDefEnum)):

                self.register_external_typedef(ext_typedef.element_typedef)

            elif isinstance(ext_typedef, types.TypeDefRecord):
                for t in ext_typedef.member_types:
                    self.register_external_typedef(t)

            elif isinstance(ext_typedef, types.TypeDefRename):
                self.register_external_typedef(ext_typedef.renamed_typedef)

            elif isinstance(ext_typedef, types.TypeDefEnum):
                self.register_external_typedef(ext_typedef.element_typedef)

            typedef = ext_typedef.copy(self.root)

            self['TypeDefinitions'].append(typedef)
            self.typedefs_by_name[typedef.type_name] = typedef
            self.typedefs_by_auid[typedef.auid] = typedef

        # make sure enum has all the values
        if isinstance(ext_typedef, (types.TypeDefExtEnum, types.TypeDefEnum)):
            ext_elements = ext_typedef.elements
            elements = typedef.elements

            for k, v in ext_elements.items():
                if k not in elements:
                    typedef.register_element(v, k)

        return typedef

    def register_external_classdef(self, ext_classdef):
        class_chain = list(ext_classdef.relatives())
        class_chain.reverse()
        for c in class_chain[1:]:
            classdef = self.lookup_classdef(c.auid)
            if not classdef:
                classdef = self.register_classdef(c.class_name,
                                                  c.auid,
                                                  c.parent.auid,
                                                  c.concrete)

            for p in c.propertydefs:
                typedef = self.register_external_typedef(p.typedef)
                pid = p.pid

                if pid >= 0x8000:
                    pid = None

                classdef.register_propertydef(p.property_name, p.auid, pid, typedef.auid, p.optional, p.unique)

        return self.lookup_classdef(ext_classdef.auid)

    def lookup_class(self, class_id):
        aaf_class = AAFClaseID_dict.get(class_id, None)
        if aaf_class:
            return aaf_class

        aaf_class = AAFClassName_dict.get(class_id, None)
        if aaf_class:
            return aaf_class

        aaf_class =  self.typedefs_classes.get(class_id, None)
        if aaf_class:
            return aaf_class

        return core.AAFObject

    def lookup_typedef(self, t):
        if isinstance(t, types.TypeDef):
            return t

        t = str2auid(t)
        if isinstance(t, AUID):
            return self.typedefs_by_auid.get(t, None)
        return self.typedefs_by_name.get(t, None)

    def lookup_classdef(self, t):
        if isinstance(t, ClassDef):
            return t
        t = str2auid(t)
        if isinstance(t, AUID):
            return self.classdefs_by_auid.get(t, None)
        return self.classdefs_by_name.get(t, None)

    @property
    def classdef(self):
        return self.classdefs_by_name['MetaDictionary']

    def next_free_pid(self):

        while self.next_pid in self.local_pids:
            self.next_pid -= 1
            if self.next_pid < 0x8000:
                raise ValueError("ran out of local pids")

        result = self.next_pid
        self.next_pid -= 1

        self.local_pids.add(result)
        return result

    def read_properties(self):
        super(MetaDictionary, self).read_properties()
        for key, typedef in self['TypeDefinitions'].items():
            self.typedefs_by_name[typedef.type_name] = typedef
            self.typedefs_by_auid[typedef.auid] = typedef

        # reset local pids
        self.local_pids = set()
        for key, classdef in self['ClassDefinitions'].items():
            self.classdefs_by_name[classdef.class_name] = classdef
            self.classdefs_by_auid[classdef.auid] = classdef
            propertydef_by_pid = {}
            for pdef in classdef['Properties'].values():
                if pdef.pid >= 0x8000:
                    self.local_pids.add(pdef.pid)

                propertydef_by_pid[pdef.pid] = pdef

            classdef.propertydef_by_pid = propertydef_by_pid


        # add typedefs not defined by file data model
        if self.root.writeable:
            for key, typedef in self.typedefs_by_auid.items():
                # skip root typedefs
                if key in (AUID('05022800-0000-0000-060E-2B3401040101'),
                           AUID('05022700-0000-0000-060E-2B3401040101')):
                    continue

                if key not in self['TypeDefinitions']:
                    # this is super annoying, handle typedefs that have dynamic pids
                    classdef = self.classdefs_by_auid.get(typedef.class_id)
                    assert classdef
                    for pdef in classdef['Properties'].values():
                        if pdef.pid >= 0x8000:
                            # if typedef has a dynamic pid might need to change it
                            if pdef.pid in self.local_pids:
                                old_pid = pdef.pid
                                new_pid = self.next_free_pid()
                                pdef.pid = new_pid

                                prop = typedef.property_entries.pop(old_pid)
                                prop.pid = new_pid
                                typedef.property_entries[new_pid] = prop

                    self['TypeDefinitions'].append(typedef)
                    if classdef.auid not in self['ClassDefinitions']:
                        self['ClassDefinitions'].append(classdef)

        # add classdefs not defined by file data model
        if self.root.writeable:
            for key, classdef in self.classdefs_by_auid.items():
                # skip root classdefs
                if key in (AUID('b3b398a5-1c90-11d4-8053-080036210804'), ):
                    continue

                if key not in self['ClassDefinitions']:

                    # handle any conflicting dynamic pids
                    for pdef in classdef['Properties'].values():
                        if pdef.pid >= 0x8000:
                            if pdef.pid in self.local_pids:
                                pdef.pid = self.next_free_pid()

                    self['ClassDefinitions'].append(classdef)



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import os
import subprocess
import unittest
import hashlib
import common
import aaf2
from aaf2 import video, audio, mobid, mobs, essence

class MXFTests(unittest.TestCase):


    def test_basic_video_link(self):
        profile_name = 'dnx_1080p_36_23.97'
        video_sample = common.generate_dnxhd(profile_name, "mxf_link_video.mxf", 10, fmt='mxf_opatom')

        meta = common.probe(video_sample)
        video_mob_id = mobid.MobID(meta['format']['tags']['material_package_umid'])
        video_source_mob_id = mobid.MobID(meta['streams'][0]['tags']['file_package_umid'])

        new_file = os.path.join(common.sandbox(), 'mxf_link_video.aaf')

        with aaf2.open(new_file, 'w') as f:
            f.content.link_external_mxf(video_sample)

        with aaf2.open(new_file, 'r') as f:
            assert len(f.content.mobs) == 2
            for mob in f.content.mobs:
                if isinstance(mob, mobs.MasterMob):
                    assert mob.mob_id == video_mob_id
                elif isinstance(mob, mobs.SourceMob):
                    if isinstance(mob.descriptor, essence.CDCIDescriptor):
                        assert mob.mob_id == video_source_mob_id

    def test_basic_audio_link(self):

        audio_profile_name = 'pcm_48000_s24le'
        sample_format = audio.pcm_profiles[audio_profile_name]['sample_format']
        sample_rate = audio.pcm_profiles[audio_profile_name]['sample_rate']
        audio_duration = 1

        audio_sample = common.generate_pcm_audio_mono(audio_profile_name, sample_format=sample_format,
                         sample_rate=sample_rate, duration=audio_duration, fmt='mxf_opatom')

        meta = common.probe(audio_sample)
        audio_mob_id = mobid.MobID(meta['format']['tags']['material_package_umid'])
        audio_source_mob_id = mobid.MobID(meta['streams'][0]['tags']['file_package_umid'])

        new_file = os.path.join(common.sandbox(), 'mxf_link_audio.aaf')

        with aaf2.open(new_file, 'w') as f:
            f.content.link_external_mxf(audio_sample)

        with aaf2.open(new_file, 'r') as f:
            assert len(f.content.mobs) == 2
            for mob in f.content.mobs:
                if isinstance(mob, mobs.MasterMob):
                    assert mob.mob_id == audio_mob_id
                elif isinstance(mob, mobs.SourceMob):
                    if isinstance(mob.descriptor, essence.PCMDescriptor):
                        assert mob.mob_id == audio_source_mob_id


if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from aaf2.auid import AUID
from uuid import UUID
import uuid
import unittest

class MobIDTests(unittest.TestCase):

    def test_basic(self):
        s = "0d010101-0101-2100-060e-2b3402060101"
        v = AUID(s)
        u = UUID(s)
        assert str(v) == s
        assert str(v.uuid) == s
        assert v.uuid == u

    def test_be(self):
        s = "0d010101-0101-2100-060e-2b3402060101"
        v = AUID(s)
        u = UUID(s)
        assert v.uuid.bytes == v.bytes_be

    def test_int(self):
        s = "0d010101-0101-2100-060e-2b3402060101"
        v = AUID(s)
        u = UUID(s)
        assert v.int == u.int

        v = AUID(int=100)
        u = UUID(int=100)
        assert v.int == u.int
        for i in range(10):
            u = uuid.uuid4()
            a = AUID(int= u.int)
            assert u.int == a.int

    def test_noargs(self):
        # expected behavour matches uuid.py
        with self.assertRaises(TypeError):
            AUID()

        # print(v.int)

if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import sys
import struct
import datetime

from io import BytesIO
import io

from .utils import (read_u8, read_u16be,
                   read_u32be, read_s32be,
                   read_u64be, read_s64be,
                   int_from_bytes)
from .mobid import MobID
from .model import datadefs
from .auid import AUID

MXF_CLASSES = {}

def register_mxf_class(classobj):
    MXF_CLASSES[classobj.class_id] = classobj
    return classobj


class MXFRef(AUID):
    pass

class MXFRefArray(list):
    pass

def read_auid_be(f):
    data = f.read(16)
    if data:
        return AUID(bytes_be=data)

def read_strongref(f):
    data = f.read(16)
    if data:
        return MXFRef(bytes_be=data)


def decode_strong_ref_array(f):

    count = read_u32be(f)
    f.read(4)
    refs = MXFRefArray()
    for i in range(count):
        refs.append(read_strongref(f))
    return refs


def decode_utf16be(data):
    return data.decode('utf-16be').split(u'\x00')[0]

def decode_auid(data):
    return AUID(bytes_be=data)

def reverse_auid(data):
    new = data.hex[16:] + data.hex[:16]
    return AUID(new)

def decode_datadef(data):
    orig = AUID(bytes_be=data)
    datadef = reverse_auid(orig)
    name = datadefs.DataDefs.get(str(datadef), (None, None))[0]
    return name

def decode_strongref(data):
    return MXFRef(bytes_be=data)

def decode_indirect_value(data):
    data = bytearray(data)
    typedef = reverse_auid(decode_auid(data[:16]))
    if typedef == AUID("00060e2b-3401-0401-4c00-021001000000"):
        assert data[16] == 0x01 # byte order?
        return data[17:].decode('utf-16le').rstrip('\x00')
    elif typedef == AUID("00060e2b-3401-0401-4201-100200000000"):
        assert data[16] == 0x01# byte order?
        return data[17:].decode('utf-16be').rstrip('\x00')
    elif typedef == AUID("00060e2b-3401-0401-4c00-070101000000"):
        assert data[16] == 0x01# byte order?
        return struct.unpack(b"<i", data[17:])[0]
    else:
        # unhandled indirect type
        return

def decode_rational(f):
    num = read_u32be(f)
    den = read_u32be(f)

    return "%d/%d" % (num, den)

def decode_video_line_map(f):
    count = read_u32be(f)
    size = read_u32be(f)
    line_map = []
    if size == 4:
        if count > 0:
            line_map.append(read_s32be(f))
        else:
            line_map.append(0)
        if count > 1:
            line_map.append(read_s32be(f))
        else:
            line_map.append(0)
    return line_map

def decode_pixel_layout(f):
    layout = []
    for i in range(8):
        code = read_u8(f)
        depth = read_u8(f)
        if not code:
            break
        layout.append({'Code':code, 'Size':depth})
    return layout

def decode_timestamp(f):
    t = read_u64be(f)

    year = t >> 48
    month = t >> 40 & 0xFF
    day = t >> 32 & 0xFF
    hour = t >> 24 & 0xFF
    minute = t >> 16 & 0xFF
    sec = t >> 8  & 0xFF
    try:
        d = datetime.datetime(year, month, day, hour, minute, sec)
        return d
    except:
        return datetime.datetime.now()


def decode_mob_id(data):
    uid1 = AUID(bytes_be=data[:16])
    uid2 =  AUID(bytes_be=data[16:])
    m = MobID(str(uid1) + str(uid2))
    return m

def ama_path(path):
    prefix ="file://"
    return prefix + path

class MXFObject(object):
    def __init__(self):
        self.instance_id = None
        self.data = {}
        self.root = None

    def create_aaf_instance():
        pass

    def read_tag(self, tag, data):
        if tag == 0x3c0a:
            self.instance_id = decode_auid(data)

    def read_properties(self, f, length, local_tags):
        for tag, data in iter_tags(f, length):
            self.read_tag(tag, data)
            uid = local_tags.get(tag, None)
            with BytesIO(data) as f:

                if uid == AUID("a0240060-94eb-75cb-ce2a-ca5051ab11d3"):
                    self.data['FrameSampleSize'] = read_s32be(f)
                elif uid == AUID("a0240060-94eb-75cb-ce2a-ca4d51ab11d3"):
                    self.data['ResolutionID'] = read_s32be(f)
                elif uid == AUID("a0220060-94eb-75cb-96c4-69924f6211d3"):
                    self.data['AppCode'] = read_s32be(f)
                elif uid == AUID("060e2b34-0101-0109-0601-010406100000"):
                    self.data['SubDescriptors'] = decode_strong_ref_array(f)
                elif uid == AUID("a01c0004-ac96-9f50-6095-818347b111d4"):
                    self.data["MobAttributeList"] = decode_strong_ref_array(f)
                elif uid == AUID("a01c0004-ac96-9f50-6095-818547b111d4"):
                    self.data["TaggedValueAttributeList"] = decode_strong_ref_array(f)

    def resolve_ref(self, key):
        ref = self.data.get(key, None)
        if ref:
            obj = self.root.resolve(ref)
            if obj:
                return obj
        raise Exception("unable to resolve: %s %s %s" % (key,auid_to_str_list(ref, sep=' '), str(ref)) )

    def iter_strong_refs(self, key):
        for ref in self.data.get(key, []):
            # print(ref, auid_to_str_list(ref, sep=' '))
            yield self.root.resolve(ref)


    def __repr__(self):
        return str(self.data)

@register_mxf_class
class MXFPreface(MXFObject):
    class_id = AUID("060e2b34-0253-0101-0d01-010101012f00")
    def read_tag(self, tag, data):
        super(MXFPreface, self).read_tag(tag, data)

        if tag == 0x3b09:
            self.data['OperationalPattern'] = decode_auid(data)
        elif tag == 0x3b03:
            self.data['ContentStorage'] = decode_strongref(data)

@register_mxf_class
class MXFContentStorage(MXFObject):
    class_id = AUID("060e2b34-0253-0101-0d01-010101011800")
    def read_tag(self, tag, data):
        super(MXFContentStorage, self).read_tag(tag, data)
        with BytesIO(data) as f:
            if tag == 0x1902:
                self.data['EssenceContainerData'] = decode_strong_ref_array(f)
            elif tag == 0x1901:
                self.data['Packages'] = decode_strong_ref_array(f)


class MXFPackage(MXFObject):

    def read_tag(self, tag, data):
        super(MXFPackage, self).read_tag(tag, data)

        with BytesIO(data) as f:

            if tag == 0x4403:
                self.data['Slots'] = decode_strong_ref_array(f)
            elif tag == 0x4401:
                self.data['MobID'] = decode_mob_id(data)
            elif tag == 0x4402:
                self.data['Name'] = decode_utf16be(data)
            elif tag == 0x4701:
                self.data['Descriptor'] = decode_strongref(data)
            elif tag == 0x4404:
                self.data['LastModified'] = decode_timestamp(f)
            elif tag == 0x4405:
                self.data['CreationTime'] = decode_timestamp(f)
            elif tag == 0x4408:
                self.data['UsageCode'] = decode_auid(data) # doesn't appear reversed...
            elif tag == 0x4406:
                self.data['UserComments'] = decode_strong_ref_array(f)

    @property
    def mob_id(self):
        return self.data.get('MobID', None)

    def link(self):
        mob = self.create_aaf_instance()

        mob.mob_id = self.mob_id
        self.root.aaf.content.mobs.append(mob)

        name = self.data.get('Name', None)
        if name:
            mob.name = name

        for track in self.iter_strong_refs("Slots"):
            # print(type(track), track)

            if isinstance(track, (MXFStaticTrack, MXFEventTrack)):
                continue

            timeline = track.link()
            mob.slots.append(timeline)

        for key in ('LastModified', 'CreationTime', 'UsageCode', 'AppCode'):
            if key in self.data:
                mob[key].value = self.data[key]

        for item in self.iter_strong_refs('UserComments'):
            tag = item.link()
            if tag:
                mob['UserComments'].append(tag)

        for item in self.iter_strong_refs('MobAttributeList'):
            tag = item.link()
            if tag:
                mob['MobAttributeList'].append(tag)

        if 'Descriptor' in self.data:
            d = self.resolve_ref('Descriptor')
            mob.descriptor = d.link()

        return mob

@register_mxf_class
class MXFMaterialPackage(MXFPackage):
    class_id = AUID("060e2b34-0253-0101-0d01-010101013600")
    def create_aaf_instance(self):
        return self.root.aaf.create.MasterMob()

@register_mxf_class
class MXFSourcePackage(MXFPackage):
    class_id = AUID("060e2b34-0253-0101-0d01-010101013700")
    def create_aaf_instance(self):
        return self.root.aaf.create.SourceMob()

@register_mxf_class
class MXFTrack(MXFObject):
    class_id = AUID("060e2b34-0253-0101-0d01-010101013b00")

    def create_aaf_instance(self):
        return self.root.aaf.create.TimelineMobSlot()

    def read_tag(self, tag, data):
        super(MXFTrack, self).read_tag(tag, data)
        with BytesIO(data) as f:
            if tag == 0x4b02:
                self.data['Origin'] = read_s64be(f)
            elif tag == 0x4b01:
                self.data['EditRate'] = decode_rational(f)
            elif tag == 0x4803:
                self.data['Segment'] = decode_strongref(data)
            elif tag == 0x4804:
                self.data['PhysicalTrackNumber'] = read_u32be(f)
            elif tag == 0x4801:
                self.data['SlotID'] = read_u32be(f)
            elif tag == 0x4802:
                self.data['SlotName'] = decode_utf16be(data)

    def link(self):
        timeline = self.create_aaf_instance()

        for key in ('SlotID', 'SlotName', 'EditRate', 'PhysicalTrackNumber', 'Origin'):
            if key in self.data:
                timeline[key].value = self.data[key]

        segment = self.resolve_ref('Segment')
        timeline.segment = segment.link()
        return timeline

@register_mxf_class
class MXFStaticTrack(MXFTrack):
    class_id = AUID("060e2b34-0253-0101-0d01-010101013a00")

    def create_aaf_instance(self):
        return self.root.aaf.create.StaticMobSlot()

@register_mxf_class
class MXFEventTrack(MXFTrack):
    class_id = AUID("060e2b34-0253-0101-0d01-010101013900")

    def create_aaf_instance(self):
        return self.root.aaf.create.EventMobSlot()

class MXFComponent(MXFObject):
    def read_tag(self, tag, data):
        super(MXFComponent, self).read_tag(tag, data)
        with BytesIO(data) as f:
            if tag == 0x1001:
                self.data['Components'] = decode_strong_ref_array(f)
            elif tag == 0x1201:
                self.data['StartTime'] = read_u64be(f)
            elif tag == 0x1102:
                self.data['SourceMobSlotID'] = read_u32be(f)
            elif tag == 0x1101:
                self.data['SourceID'] = decode_mob_id(data)
            elif tag == 0x0202:
                self.data['Length'] = read_u64be(f)
            elif tag == 0x0201:
                self.data['DataDef'] = decode_datadef(data)
            elif tag == 0x1503:
                self.data['Drop'] = read_u8(f) == 1
            elif tag == 0x1502:
                self.data['FPS'] = read_u16be(f)
            elif tag == 0x1501:
                self.data['Start'] = read_u64be(f)
            elif tag == 0x0501:
                self.data['Choices'] = decode_strong_ref_array(f)
            elif tag == 0x0502:
                self.data['StillFrame'] = decode_strongref(data)
            elif tag == 0x0d01:
                self.data['InputSegment'] = decode_strongref(data)
            elif tag == 0x0d02:
                self.data['PulldownKind'] = read_u8(f)
            elif tag == 0x0d03:
                self.data['PulldownDirection'] = read_u8(f)
            elif tag == 0x0d04:
                self.data['PhaseFrame'] = read_s32be(f)
            elif tag == 0x0e01:
                self.data['RelativeScope'] = read_s32be(f)
            elif tag == 0x0e02:
                self.data['RelativeSlot'] = read_s32be(f)

@register_mxf_class
class MXFSequence(MXFComponent):
    class_id = AUID("060e2b34-0253-0101-0d01-010101010f00")

    def create_aaf_instance(self):
        return self.root.aaf.create.Sequence()

    def link(self):
        s = self.create_aaf_instance()

        s.media_kind = self.data['DataDef'] #or 'DataDef_Unknown'
        s.length = self.data.get('Length', 0)
        # for item in self.data['Components']:
        #     print(auid_to_str_list(item, sep=' '), item)
        for item in self.iter_strong_refs('Components'):
            s['Components'].append(item.link())
        return s

@register_mxf_class
class MXFSourceClip(MXFComponent):
    class_id = AUID("060e2b34-0253-0101-0d01-010101011100")

    def create_aaf_instance(self):
        return self.root.aaf.create.SourceClip()

    def link(self):
        s = self.create_aaf_instance()
        s.media_kind = self.data['DataDef'] or 'DataDef_Unknown'
        for key in ('SourceID', 'SourceMobSlotID', 'StartTime', 'Length'):
            s[key].value = self.data.get(key, None)
        return s

@register_mxf_class
class MXFTimecode(MXFComponent):
    class_id = AUID("060e2b34-0253-0101-0d01-010101011400")

    def create_aaf_instance(self):
        return self.root.aaf.create.Timecode()

    def link(self):

        s = self.create_aaf_instance()
        for key in ('FPS', 'Drop', 'Start'):
            s[key].value = self.data[key]

        s.media_kind = self.data['DataDef']
        s.length = self.data['Length']

        return s

@register_mxf_class
class MXFPulldown(MXFComponent):
    class_id = AUID("060e2b34-0253-0101-0d01-010101010c00")

    def create_aaf_instance(self):
        return self.root.aaf.create.Pulldown()

    def link(self):

        p = self.create_aaf_instance()
        p.media_kind = self.data['DataDef']
        p.length = self.data['Length']

        p['InputSegment'].value = self.resolve_ref('InputSegment').link()

        for key in ('PhaseFrame', 'PulldownDirection', 'PulldownKind'):
            p[key].value = self.data[key]
        return p

@register_mxf_class
class MXFFiller(MXFComponent):
    class_id = AUID("060e2b34-0253-0101-0d01-010101010900")

    def create_aaf_instance(self):
        return self.root.aaf.create.Filler()

    def link(self):
        c = self.create_aaf_instance()
        c.media_kind = self.data['DataDef']
        c.length = self.data['Length']
        return c

@register_mxf_class
class MXFScopeReference(MXFComponent):
    class_id = AUID("060e2b34-0253-0101-0d01-010101010d00")

    def create_aaf_instance(self):
        return self.root.aaf.create.ScopeReference()

    def link(self):
        c = self.create_aaf_instance()
        c.media_kind = self.data['DataDef']
        c.length = self.data['Length']
        for key in ('RelativeSlot', 'RelativeScope'):
            c[key].value = self.data[key]

        return c

@register_mxf_class
class MXFEssenceGroup(MXFComponent):
    class_id = AUID("060e2b34-0253-0101-0d01-010101010500")

    def create_aaf_instance(self):
        return self.root.aaf.create.EssenceGroup()

    def link(self):

        e = self.create_aaf_instance()
        e.media_kind = self.data['DataDef']
        e.length = self.data['Length']
        e['Choices'].value = [item.link() for item in self.iter_strong_refs('Choices')]

        return e

class MXFDescriptor(MXFObject):

    def read_tag(self, tag, data):
        super(MXFDescriptor, self).read_tag(tag, data)
        with BytesIO(data) as f:
            if tag == 0x3f01:
                self.data['FileDescriptors'] = decode_strong_ref_array(f)
            elif tag == 0x3004:
                self.data['ContainerFormat'] = reverse_auid(decode_auid(data))
            elif tag == 0x3005:
                self.data['CodecDefinition'] = reverse_auid(decode_auid(data))
            elif tag == 0x3006:
                self.data['LinkedSlotID'] = read_u32be(f)
            elif tag == 0x3203:
                self.data['StoredWidth'] = read_u32be(f)
            elif tag == 0x3202:
                self.data['StoredHeight'] = read_u32be(f)
            elif tag == 0x3208: #this is display
                self.data['SampledHeight'] = read_u32be(f)
            elif tag == 0x3209: #this is display
                self.data['SampledWidth'] = read_u32be(f)
            elif tag == 0x320d:
                self.data['VideoLineMap'] = decode_video_line_map(f)
            elif tag == 0x3211:
                self.data['ImageAlignmentOffset'] = read_u32be(f)
            elif tag == 0x3002:
                self.data['Length'] = read_s64be(f)
            elif tag == 0x3001:
                self.data['SampleRate'] = decode_rational(f)
            elif tag == 0x3d03:
                self.data['AudioSamplingRate'] = decode_rational(f)
            elif tag == 0x3d0a:
                self.data['BlockAlign'] = read_u16be(f)
            elif tag == 0x3d01:
                self.data['QuantizationBits'] = read_u32be(f)
            elif tag == 0x3d07:
                self.data['Channels'] = read_u32be(f)
            elif tag == 0x3d09:
                self.data['AverageBPS'] = read_u32be(f)
            elif tag == 0x3d02:
                self.data['Locked'] = read_u8(f) == 1
            elif tag == 0x3301:
                self.data['ComponentWidth'] = read_u32be(f)
            elif tag == 0x320c:
                self.data['FrameLayout'] = read_u8(f)
            elif tag == 0x320e:
                self.data['ImageAspectRatio'] = decode_rational(f)
            elif tag == 0x3d06:
                self.data['SoundCompression'] =  reverse_auid(decode_auid(data))
            elif tag == 0x3201:
                self.data['Compression'] = reverse_auid(decode_auid(data))
            elif tag == 0x3302:
                self.data['HorizontalSubsampling'] = read_u32be(f)
            elif tag == 0x3308:
                self.data['VerticalSubsampling'] = read_u32be(f)
            elif tag == 0x2f01:
                self.data['Locator'] = decode_strong_ref_array(f)
            elif tag == 0x3401:
                self.data['PixelLayout'] = decode_pixel_layout(f)

@register_mxf_class
class MXFMultipleDescriptor(MXFDescriptor):
    class_id = AUID("060e2b34-0253-0101-0d01-010101014400")

    def create_aaf_instance(self):
        return self.root.aaf.create.MultipleDescriptor()

    def link(self):
        d = self.create_aaf_instance()
        for item in self.iter_strong_refs("FileDescriptors"):
            # if isinstance(item, MXFAES3AudioDescriptor):
            #     continue
            d['FileDescriptors'].append(item.link())
        d['Length'].value = self.data.get('Length', 0)

        n = self.root.aaf.create.NetworkLocator()
        n['URLString'].value = ama_path(self.root.path)
        d['Locator'].append(n)

        d['SampleRate'].value = self.data['SampleRate']

        if self.root.ama:
            n = self.root.aaf.create.NetworkLocator()
            n['URLString'].value = ama_path(self.root.path)
            d['Locator'].append(n)
            d['MediaContainerGUID'].value = AUID("60eb8921-2a02-4406-891c-d9b6a6ae0645")

        return d

@register_mxf_class
class MXFCDCIDescriptor(MXFDescriptor):
    class_id = AUID("060e2b34-0253-0101-0d01-010101012800")

    def create_aaf_instance(self):
        return self.root.aaf.create.CDCIDescriptor()

    def link(self):
        d = self.create_aaf_instance()

        # required
        for key in ('ComponentWidth', 'HorizontalSubsampling', 'ImageAspectRatio',
           'StoredWidth', 'VideoLineMap', 'StoredHeight', 'SampleRate', 'FrameLayout'):
           d[key].value = self.data[key]

        d['Length'].value = self.data.get('Length', 0)

        # optional
        for key in ('FrameSampleSize', 'ResolutionID', 'Compression', 'VerticalSubsampling',
                    'SampledWidth', 'SampledHeight', 'LinkedSlotID'):
            if key in self.data:
                d[key].value = self.data[key]

        for item in self.iter_strong_refs("Locator"):
            d['Locator'].append(item.link())
            n = self.root.aaf.create.NetworkLocator()
            n['URLString'].value = ama_path(self.root.path)
            d['Locator'].append(n)

        d['ContainerFormat'].value = self.root.aaf.dictionary.lookup_containerdef("AAFKLV")
        if self.root.ama:
            n = self.root.aaf.create.NetworkLocator()
            n['URLString'].value = ama_path(self.root.path)
            d['Locator'].append(n)
            d['MediaContainerGUID'].value = AUID("60eb8921-2a02-4406-891c-d9b6a6ae0645")

        return d

@register_mxf_class
class MXFRGBADescriptor(MXFDescriptor):
    class_id = AUID("060e2b34-0253-0101-0d01-010101012900")

    def create_aaf_instance(self):
        return self.root.aaf.create.RGBADescriptor()

    def link(self):
        d = self.create_aaf_instance()

        for key in ('ImageAspectRatio', 'StoredWidth', 'FrameLayout', 'PixelLayout',
                    'VideoLineMap', 'StoredHeight', 'SampleRate', ):
            d[key].value = self.data[key]

        d['Length'].value = self.data.get('Length', 0)

        # optional
        for key in ('FrameSampleSize','SampledWidth', 'SampledHeight', 'Compression', 'LinkedSlotID'):
            if key in self.data:
                d[key].value = self.data[key]

        d['ContainerFormat'].value = self.root.aaf.dictionary.lookup_containerdef("AAFKLV")
        n = self.root.aaf.create.NetworkLocator()
        n['URLString'].value = ama_path(self.root.path)
        d['Locator'].append(n)
        if self.root.ama:
            n = self.root.aaf.create.NetworkLocator()
            n['URLString'].value = ama_path(self.root.path)
            d['Locator'].append(n)
            d['MediaContainerGUID'].value = AUID("60eb8921-2a02-4406-891c-d9b6a6ae0645")


        return d

@register_mxf_class
class MXFANCDataDescriptor(MXFDescriptor):
    class_id = AUID("060e2b34-0253-0101-0d01-010101015c00")

    def create_aaf_instance(self):
        return self.root.aaf.create.ANCDataDescriptor()

    def link(self):
        d = self.create_aaf_instance()
        for key in ('SampleRate', ):
            d[key].value = self.data[key]

        d['Length'].value = self.data.get('Length', 0)

        # optional
        for key in ('LinkedSlotID', ):
            if key in self.data:
                d[key].value = self.data[key]

        return d

@register_mxf_class
class MXFMPEG2VideoDescriptor(MXFCDCIDescriptor):
    class_id = AUID("060e2b34-0253-0101-0d01-010101015100")

    def link(self):
        # 060e2b34.04010103.04010202.01040300
        self.data['ResolutionID'] = 4076 #XDCAM HD 50Mbit

        d['Length'].value = self.data.get('Length', 0)

        # optional
        for key in ('LinkedSlotID', ):
            if key in self.data:
                d[key].value = self.data[key]

        return super(MXFMPEG2VideoDescriptor, self).link()

@register_mxf_class
class MXFPCMDescriptor(MXFDescriptor):
    class_id = AUID("060e2b34-0253-0101-0d01-010101014800")

    def create_aaf_instance(self):
        return self.root.aaf.create.PCMDescriptor()

    def link(self):
        d = self.create_aaf_instance()
        # required
        for key in ('BlockAlign', 'AverageBPS', 'Channels',
            'QuantizationBits', 'AudioSamplingRate', 'SampleRate'):
            d[key].value = self.data[key]

        d['Length'].value = self.data.get('Length', 0)

        # optional
        for key in ('LinkedSlotID', ):
            if key in self.data:
                d[key].value = self.data[key]

        n = self.root.aaf.create.NetworkLocator()
        n['URLString'].value = ama_path(self.root.path)
        d['Locator'].append(n)
        if self.root.ama:
            n =  self.root.aaf.create.NetworkLocator()
            n['URLString'].value = ama_path(self.root.path)
            d['Locator'].append(n)
            d['MediaContainerGUID'].value = AUID("60eb8921-2a02-4406-891c-d9b6a6ae0645")

        return d

@register_mxf_class
class MXFAES3AudioDescriptor(MXFPCMDescriptor):
    class_id = AUID("060e2b34-0253-0101-0d01-010101014700")

@register_mxf_class
class MXFSoundDescriptor(MXFPCMDescriptor):
    class_id = AUID("060e2b34-0253-0101-0d01-010101014200")

@register_mxf_class
class MXFImportDescriptor(MXFDescriptor):
    class_id = AUID("060e2b34-0253-0101-0d01-010101014a00")

    def create_aaf_instance(self):
        return self.root.aaf.create.ImportDescriptor()

    def link(self):
        d = self.create_aaf_instance()
        n = self.root.aaf.create.NetworkLocator()
        n['URLString'].value = ama_path(self.root.path)
        d['Locator'].append(n)

        return d

@register_mxf_class
class MXFTapeDescriptor(MXFDescriptor):
    class_id = AUID("060e2b34-0253-0101-0d01-010101012e00")

    def create_aaf_instance(self):
        return self.root.aaf.create.TapeDescriptor()

    def link(self):
        return self.create_aaf_instance()

class MXFLocator(MXFObject):

    def read_tag(self, tag, data):
        super(MXFLocator, self).read_tag(tag, data)

        if tag == 0x4001:
            self.data['URLString'] =  decode_utf16be(data)

@register_mxf_class
class MXFNetworkLocator(MXFLocator):
    class_id = AUID("060e2b34-0253-0101-0d01-010101013200")

    def create_aaf_instance(self):
        return self.root.aaf.create.NetworkLocator()

    def link(self):
        n = self.create_aaf_instance()
        n['URLString'].value = self.data['URLString']
        return n

@register_mxf_class
class MXFEssenceData(MXFObject):
    class_id = AUID("060e2b34-0253-0101-0d01-010101012300")
    def read_tag(self, tag, data):
        super(MXFEssenceData, self).read_tag(tag, data)

        if tag == 0x2701:
            self.data['MobID'] = decode_mob_id(data)

@register_mxf_class
class MXFTaggedValue(MXFObject):
    class_id = AUID("060e2b34-0253-0101-0d01-010101013f00")

    def create_aaf_instance(self):
        return self.root.aaf.create.TaggedValue()

    def read_tag(self, tag, data):
        super(MXFTaggedValue, self).read_tag(tag, data)

        if tag == 0x5001:
            self.data['Name'] = decode_utf16be(data)
        elif tag == 0x5003:
            self.data['Value'] = decode_indirect_value(data)

    def link(self):
        v = self.data.get("Value", None)
        if v is None:
            return

        tag = self.create_aaf_instance()
        tag.name = self.data["Name"]
        tag.value = v

        for item in self.iter_strong_refs('TaggedValueAttributeList'):
            attr = item.link()
            if attr:
                tag['TaggedValueAttributeList'].append(attr)

        return tag

def ber_length(f):
    length = read_u8(f)
    if length > 127:
        data = bytearray(length - 128)
        bytes_read = f.readinto(data)
        assert bytes_read == len(data)
        length = int_from_bytes(data, byte_order='big')
    return length


def iter_kl(f):
    pos = f.tell()
    while True:
        # read the key
        f.seek(pos)
        key = read_auid_be(f)
        if not key:
            break
        # read the ber_length
        length = ber_length(f)

        pos = f.tell() + length
        yield key, length

def iter_tags(f, length):
    while length > 0:
        tag = read_u16be(f)
        size = read_u16be(f)
        # print("   tag 0x%04x %d" % (tag, size))
        if size:
            yield tag, f.read(size)
        length -= 4 + size

def auid_to_str_list(v, sep=',', prefix=""):
    return sep.join('%s%02x' % (prefix, i)  for i in bytearray(v.bytes_be))

class MXFFile(object):

    def __init__(self, path):
        self.objects = {}
        self.local_tags = {}
        self.preface = None
        self.header_operation_pattern = None
        self.header_partition_size = None
        self.path = path
        self.ama = False
        self.aaf = None
        with io.open(path, 'rb') as f:

            for key, length in iter_kl(f):

                # only read until the first body partition
                if not self.header_partition_size is None and f.tell() > self.header_partition_size:
                    break

                if key == AUID("060e2b34-0205-0101-0d01-020101050100"):
                    self.local_tags = self.read_primer(f, length)
                elif key == AUID("060e2b34-0205-0101-0d01-020101020400"):
                    self.read_header(f, length)
                else:
                    # print('{')
                    # print(key,  auid_to_str_list(key, sep='.'))
                    obj = self.read_object(f, key, length)
                    if obj:
                        # print(obj.__class__.__name__, obj.instance_id)
                        obj.root = self
                        self.objects[obj.instance_id] = obj

                    if isinstance(obj, MXFPreface):
                        self.preface = obj
                    # print('}')

    def resolve(self, ref):
        if isinstance(ref, MXFRef):
            return self.objects.get(ref)

    @property
    def content(self):
        return self.preface.resolve_ref("ContentStorage")

    def packages(self):
        if self.content:
            for package in self.content.iter_strong_refs("Packages"):
                yield package

    def material_packages(self):
        for package in self.packages():
            if isinstance(package, MXFMaterialPackage):
                yield package

    def link(self, f):
        self.aaf = f
        mobs = []
        for package in self.packages():
            if package.mob_id in f.content.mobs:
                continue

            mobs.append(package.link())

        return mobs

    def round_to_kag(self, pos, kag_size):
        ret = (pos // kag_size) * kag_size
        if ret == pos:
            return pos
        return pos + kag_size


    def read_header(self, f, length):
        major_version = read_u16be(f)
        minor_version = read_u16be(f)
        kag_size = read_u32be(f)
        this_partion = read_u64be(f)
        prev_partion = read_u64be(f)
        footer_partion = read_u64be(f)
        header_byte_count = read_u64be(f)
        index_byte_count = read_u64be(f)
        index_sid = read_u32be(f)
        body_offset = read_u64be(f)
        body_sid = read_u32be(f)
        self.header_operation_pattern = read_auid_be(f)

        self.header_partition_size = (self.round_to_kag(length, kag_size) +
                                      self.round_to_kag(header_byte_count, kag_size) +
                                      self.round_to_kag(index_byte_count, kag_size))

    def read_primer(self, f, length):

        item_num = read_u32be(f)
        item_len = read_u32be(f)
        if item_len != 18:
            return
        if item_num > 65536:
            return

        tags = {}
        for i in range(item_num):
            tag = read_u16be(f)
            uid = read_auid_be(f)
            # print("%04x" % tag, ':', uid)
            tags[tag] = uid

        return tags

    def read_object(self, f, key, length):

        b = bytearray(key.bytes_be)
        if not b[5] == 0x53:
            return

        obj_class = MXF_CLASSES.get(key, None)
        if obj_class:

            obj = obj_class()
            obj.root = self
            obj.read_properties(f, length, self.local_tags)
            return obj
        else:
            for tag, data in iter_tags(f, length):
                pass

    def dump_flat(self):
        for key,value in self.objects.items():
            print(value.__class__.__name__, key)
            for p, v in value.data.items():
                print("  ",p, v)

    def dump(self, obj=None, space=""):
        if obj is None:
            obj = self.preface

        print (space, obj.__class__.__name__, obj.instance_id)

        space += " "
        for key, value in sorted(obj.data.items()):
            if isinstance(value, MXFRef):
                c = self.objects.get(value, None)
                if c:
                    self.dump(c, space)
                else:
                    print(space, None)

            elif isinstance(value, MXFRefArray):
                print (space, key)
                for item in value:
                    c = self.objects.get(item, None)
                    if c:
                        self.dump(c, space + " ")
                    else:
                        print(space, None)
            else:
                print (space, key, value)

    @property
    def operation_pattern(self):

        if self.header_operation_pattern:
            op = self.header_operation_pattern
        else:
            op = self.preface.data.get('OperationalPattern', None)

        if not op:
            return

        op = bytearray(op.bytes_be)

        prefix1 = bytearray([0x06, 0x0e, 0x2b, 0x34, 0x04, 0x01, 0x01, 0x01, 0x0d, 0x01, 0x02, 0x01])
        prefix2 = bytearray([0x06, 0x0e, 0x2b, 0x34, 0x04, 0x01, 0x01, 0x02, 0x0d, 0x01, 0x02, 0x01])
        prefix3 = bytearray([0x06, 0x0e, 0x2b, 0x34, 0x04, 0x01, 0x01, 0x03, 0x0d, 0x01, 0x02, 0x01])

        prefix_valid = False
        for prefix in (prefix1, prefix2, prefix3):
            if op[:len(prefix)] == prefix:
                prefix_valid = True
                break

        if not prefix_valid:
            return

        complexity = op[12]

        if complexity >= 1 and complexity <= 3:
            package_complexity = op[13]
            letter = {1:'a', 2:'b', 3:'c'}.get(package_complexity, None)
            if letter:
                return 'OP%d%s' % (complexity, letter)

        elif complexity >= 0x10 and complexity <= 0x7f:
            if complexity == 0x10:
                return 'OPAtom'



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import aaf2
import os
import io
import time
import common
import unittest

DIFSECT    = 0xFFFFFFFC
FATSECT    = 0xFFFFFFFD
ENDOFCHAIN = 0xFFFFFFFE
FREESECT   = 0xFFFFFFFF

RANGELOCKSECT = (0x7FFFFF00 // 4096) - 1
class ImportTests(unittest.TestCase):
    def test_range_lock(self):

        frame_rate = '23.97'
        size = (960, 540)
        profile_name = 'dnxhr_lb'
        frames = 1600
        overwrite = True
        sample = common.generate_dnxhd(profile_name, "%s-import.dnxhd" % profile_name,
               frames=frames, size=size, frame_rate=frame_rate, overwrite=overwrite)

        test_file = common.get_test_file("test-rangelock.aaf")
        with aaf2.open(test_file, 'wb+') as f:
            pass
        mobids = []
        s =time.time()

        # generate a aaf file thats larger then 2gb
        with aaf2.open(test_file, 'rb+') as f:
            for v in range(30):
                start = time.time()
                mob = f.create.MasterMob("item-%d" % v)
                f.content.mobs.append(mob)
                mob.import_dnxhd_essence(sample, frame_rate)

                # print(v, time.time() - start)

        print('%d secs' % (time.time() - s))
        with aaf2.open(test_file, 'r') as f:
            # print(RANGELOCKSECT)
            # print(len(f.cfb.fat))
            # print(f.cfb.fat[RANGELOCKSECT])
            assert f.cfb.fat[RANGELOCKSECT] == ENDOFCHAIN

            # for m in f.content.essencedata:
            #     print(m)
            # for m in mobids:
            #     print(f.content.essencedata.get(m))


if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import wave
import struct

pcm_profiles = {
'pcm_32000_s16le' : {'sample_format' : 'pcm_s16le', 'sample_rate' : 32000},
'pcm_32000_s24le' : {'sample_format' : 'pcm_s24le', 'sample_rate' : 32000},

'pcm_44100_s16le' : {'sample_format' : 'pcm_s16le', 'sample_rate' : 44100},
'pcm_44100_s24le' : {'sample_format' : 'pcm_s24le', 'sample_rate' : 44100},

'pcm_48000_s16le' : {'sample_format' : 'pcm_s16le', 'sample_rate' : 48000},
'pcm_48000_s24le' : {'sample_format' : 'pcm_s24le', 'sample_rate' : 48000},
}

audio_format_sizes = {
's16'   :  (16, 2),
's32'   :  (32, 4),
'flt'   :  (32, 4),
'dbl'   :  (64, 8),
'u8p'   :  (8,  1),
's16p'  :  (16, 2),
's32p'  :  (32, 4),
'fltp'  :  (32, 4),
'dblp'  :  (64, 8),
's64'   :  (64, 8),
's64p'  :  (64, 8),
}

WAVE_EXTENSIBLE_PCM=0xFFFE


class WaveReader(wave.Wave_read):
    def __init__(self, f):
        self._blockalign = None
        # can't use super in OldStyle 2.7 class
        wave.Wave_read.__init__(self, f)

    def _read_fmt_chunk(self, chunk):

        # added support for wave extensible

        wFormatTag, self._nchannels, self._framerate, dwAvgBytesPerSec, wBlockAlign = struct.unpack_from(b'<HHLLH', chunk.read(14))
        if wFormatTag in  (wave.WAVE_FORMAT_PCM, WAVE_EXTENSIBLE_PCM):
            sampwidth = struct.unpack_from(b'<H', chunk.read(2))[0]
            self._sampwidth = (sampwidth + 7) // 8
        else:
            raise wave.Error('unknown format: %r' % (wFormatTag,))

        if not self._sampwidth in (2, 3):
            raise wave.Error('unsupported sample width: %r' % (self._sampwidth,))

        self._framesize = self._nchannels * self._sampwidth
        self._comptype = 'NONE'
        self._blockalign = wBlockAlign
        self._compname = 'not compressed'

    def getblockalign(self):
        return self._blockalign



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from aaf2.cfb import CompoundFileBinary
import os
import io

base = os.path.join(os.path.dirname(os.path.abspath(__file__)))
test_dir = os.path.join(base, 'results')
if not os.path.exists(test_dir):
    os.makedirs(test_dir)

test_files = os.path.join(base, 'test_files')

import unittest

class CopyTests(unittest.TestCase):
    def test_copy(self):
        src_path = os.path.join(test_files, "test_file_01.aaf")
        dst_path = os.path.join(test_dir, "test_copy.aaf")

        with io.open(src_path, 'rb') as file_a:
            ss_a = CompoundFileBinary(file_a, 'rb')
            with io.open(dst_path, 'wb+') as file_b:
                ss_b = CompoundFileBinary(file_b, 'wb+')

                # copy everything
                for root, storage, streams in ss_a.walk():
                    for item in storage:
                        entry = ss_b.makedir(item.path(), class_id=item.class_id)

                    for item in streams:
                        s_a = ss_a.open(item.path(), 'r')
                        s_b = ss_b.open(item.path(), 'w')
                        s_b.write(s_a.read())

                # check everything exists while file live
                for root, storage, streams in ss_a.walk():

                    assert ss_b.exists(root.path())
                    for item in storage:
                        assert ss_b.exists(item.path())
                        s_b = ss_b.find(item.path())
                        assert s_b.class_id == item.class_id

                    for item in streams:
                        s_a = ss_a.open(item.path(), 'r')
                        s_b = ss_b.open(item.path(), 'r')
                        assert s_a.read() == s_b.read()

                ss_b.close()

            # reopen file and check everything exists
            with io.open(dst_path, 'rb') as f:
                ss_b = CompoundFileBinary(f, 'rb')
                # check everything exists
                for root, storage, streams in ss_a.walk():

                    assert ss_b.exists(root.path())
                    for item in storage:
                        assert ss_b.exists(item.path())
                        s_b = ss_b.find(item.path())
                        assert s_b.class_id == item.class_id

                    for item in streams:
                        s_a = ss_a.open(item.path(), 'r')
                        s_b = ss_b.open(item.path(), 'r')
                        assert s_a.read() == s_b.read()


if __name__ == '__main__':

    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from aaf2.cfb import CompoundFileBinary

from struct import unpack
import os
base = os.path.join(os.path.dirname(os.path.abspath(__file__)))
test_dir = os.path.join(base, 'results')
if not os.path.exists(test_dir):
    os.makedirs(test_dir)

test_files = os.path.join(base, 'test_files')


if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.DEBUG)
    test_file = os.path.join(test_dir, "write_test.aaf")
    f = open(test_file, 'wb+')

    ss = CompoundFileBinary(f)
    # f.flush()

    # ss.read_header()
    # ss.debug_grow = True
    try:
        for i in range(100):
            entry = ss.create_dir_entry("/test%d" % i, 'storage')
            entry = ss.create_dir_entry("/test%d/test" % i, 'storage')
            entry = ss.create_dir_entry("/test%d/test2" % i, 'storage')
            # print ss.next_free_dir_id()

        ss.remove("/test99/test2")
        ss.remove("/test99/test")
        ss.remove("/test99")

        ss.remove("/test97/test")

        ss.rmtree("test87")
    finally:
        ss.close()
        f.close()

    f = open(test_file, 'rb')
    ss = CompoundFileBinary(f)

    for root, storage, streams in ss.walk():
        # print root.path()
        for item in storage:
            print(item.path())

        # for item in streams:
        #     # data = item.data()
        #     # print item
        #     print item.path()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import unittest
import aaf2
import common
import uuid


class DictionaryTests(unittest.TestCase):

    def test_operationdef(self):

        result_file = common.get_test_file('operation_def.aaf')

        parm_id = uuid.uuid4()
        effect_id = uuid.uuid4()

        with aaf2.open(result_file, 'w') as f:

            typedef = f.dictionary.lookup_typedef("Rational")
            param = f.create.ParameterDef(parm_id, "AvidEffectID", "avid effect id", typedef)
            f.dictionary.register_def(param)

            op_def = f.create.OperationDef(effect_id, "MatteKey_2", "matte key def")
            op_def.media_kind = "Picture"
            f.dictionary.register_def(op_def)
            op_def['IsTimeWarp'].value = False
            op_def['Bypass'].value = 2
            op_def['NumberInputs'].value = 3
            op_def['OperationCategory'].value = "OperationCategory_Effect"
            op_def['ParametersDefined'].append(param)

            length = 100
            operation = f.create.OperationGroup(op_def, length)

            mob = f.create.MasterMob()
            f.content.mobs.append(mob)
            slot = mob.create_picture_slot()
            slot.segment.components.append(operation)

        with aaf2.open(result_file, 'r') as f:
            param = f.dictionary.lookup_parameterdef('AvidEffectID')
            assert param.auid == parm_id

            op_def = f.dictionary.lookup_operationdef("MatteKey_2")
            assert op_def.auid == effect_id
            assert op_def.media_kind == 'Picture'
            assert op_def['IsTimeWarp'].value == False
            assert op_def['Bypass'].value == 2
            assert op_def['NumberInputs'].value == 3
            assert op_def['OperationCategory'].value == "OperationCategory_Effect"

            assert param in op_def['ParametersDefined'].value

            mob = next(f.content.mobs.values())
            slot = mob.slots.value[0]
            op = slot.segment.components.value[0]
            assert op.operation == op_def

    def test_interpolationdef(self):
        result_file = common.get_test_file('interpolation_def.aaf')

        interpolation_id = uuid.uuid4()
        name = "TestInterp"
        with aaf2.open(result_file, 'w') as f:

            interpolation_id =interpolation_id
            interpolationdef = f.create.InterpolationDef(interpolation_id, name, "interpolation")
            f.dictionary.register_def(interpolationdef)

            varying_value = f.create.VaryingValue()
            varying_value['Interpolation'].value = interpolationdef

            assert interpolationdef is f.dictionary.lookup_interperlationdef(name)

        with aaf2.open(result_file, 'r') as f:
             interpolationdef = f.dictionary.lookup_interperlationdef(name)

             assert interpolationdef.auid == interpolation_id
             assert interpolationdef.name == name

if __name__ == "__main__":
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import traceback

from . import core
from . mobid import MobID
from .utils import register_class
from .auid import AUID

@register_class
class EssenceData(core.AAFObject):
    class_id = AUID("0d010101-0101-2300-060e-2b3402060101")
    __slots__ = ()

    @property
    def unique_key(self):
        return self.mob_id

    @property
    def mob_id(self):
        return self['MobID'].value

    @mob_id.setter
    def mob_id(self, value):
        self['MobID'].value = value

    @property
    def mob(self):
        mob_id = self.mob_id
        if mob_id:
            return self.root.content.mobs.get(mob_id, None)
        return None

    @mob.setter
    def mob(self, value):
        self.mob_id = value.mob_id

    def open(self, mode='r'):
        return self['Data'].open(mode)

@register_class
class EssenceDescriptor(core.AAFObject):
    class_id = AUID("0d010101-0101-2400-060e-2b3402060101")
    __slots__ = ()

    @property
    def locator(self):
        return self['Locator'].value

@register_class
class FileDescriptor(EssenceDescriptor):
    class_id = AUID("0d010101-0101-2500-060e-2b3402060101")
    __slots__ = ()

    @property
    def length(self):
        return self['Length'].value
    @length.setter
    def length(self, value):
        self['Length'].value = value

@register_class
class DigitalImageDescriptor(FileDescriptor):
    class_id = AUID("0d010101-0101-2700-060e-2b3402060101")
    __slots__ = ()

@register_class
class CDCIDescriptor(DigitalImageDescriptor):
    class_id = AUID("0d010101-0101-2800-060e-2b3402060101")
    __slots__ = ()

@register_class
class RGBADescriptor(DigitalImageDescriptor):
    class_id = AUID("0d010101-0101-2900-060e-2b3402060101")
    __slots__ = ()

    @property
    def pixel_layout(self):
        return self['PixelLayout'].value

@register_class
class TapeDescriptor(EssenceDescriptor):
    class_id = AUID("0d010101-0101-2e00-060e-2b3402060101")
    __slots__ = ()

@register_class
class SoundDescriptor(FileDescriptor):
    class_id = AUID("0d010101-0101-4200-060e-2b3402060101")
    __slots__ = ()

@register_class
class WAVEDescriptor(FileDescriptor):
    # from ... https://github.com/ebu/ebu-libmxf/blob/master/tools/MXFDump/AAFMetaDictionary.h
    class_id = AUID("0d010101-0101-2c00-060e-2b3402060101")
    __slots__ = ()

@register_class
class AIFCDescriptor(FileDescriptor):
    class_id = AUID("0d010101-0101-2600-060e-2b3402060101")
    __slots__ = ()

@register_class
class DataEssenceDescriptor(FileDescriptor):
    class_id = AUID("0d010101-0101-4300-060e-2b3402060101")
    __slots__ = ()

@register_class
class MultipleDescriptor(FileDescriptor):
    class_id = AUID("0d010101-0101-4400-060e-2b3402060101")
    __slots__ = ()

@register_class
class PCMDescriptor(SoundDescriptor):
    class_id = AUID("0d010101-0101-4800-060e-2b3402060101")
    __slots__ = ()

@register_class
class PhysicalDescriptor(EssenceDescriptor):
    class_id = AUID("0d010101-0101-4900-060e-2b3402060101")
    __slots__ = ()

@register_class
class ImportDescriptor(PhysicalDescriptor):
    class_id = AUID("0d010101-0101-4a00-060e-2b3402060101")
    __slots__ = ()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from aaf2 import properties


def test_mangle():

    for item, pid, size in (('tessdsdsddssssssssssttwtestestst',1000, 32),
                            ('tesstwtestst',1000, 32),
                            ('TaggedValueAttributeList', 0xffd0, 32-10)):

        result = properties.mangle_name(item, pid, size)
        assert len(result) <= size


if __name__ == "__main__":
    test_mangle()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import aaf2
from uuid import UUID
import uuid
import unittest
import common
import shutil
from aaf2.auid import AUID

def has_duplicate_pid(f):
    pids = []
    for classdef in f.metadict['ClassDefinitions'].values():
        for pdef in classdef['Properties'].values():
            pid = pdef['LocalIdentification'].value
            if pid in pids:
                return True
            pids.append(pid)
    return False

class ModelTests(unittest.TestCase):

    def test_no_root_class(self):

        # make sure we don't write the root_class definition
        result_file = common.get_test_file('root_tess.aaf')
        with aaf2.open(result_file, 'w') as f:
            pass
        root_auid = AUID('b3b398a5-1c90-11d4-8053-080036210804')
        with aaf2.open(result_file, 'r') as f:
            for classdef in f.metadict['ClassDefinitions'].value:
                self.assertFalse(classdef.class_name == "Root")
                self.assertFalse(classdef.auid == root_auid)
                self.assertFalse(classdef.uuid == UUID(bytes_le=bytes(root_auid.bytes_le)))

    def test_register(self):

        result_file = common.get_test_file('register_class.aaf')

        test_classdef_auid = AUID(int=1)
        test_prop_auid = AUID(int=42)
        test_prop_pid = 0xBEEF
        parent_class_name = 'EssenceDescriptor'
        with aaf2.open(result_file, 'w') as f:
            test_classdef = f.metadict.register_classdef("TestClass", test_classdef_auid, parent_class_name, True)
            assert isinstance(test_classdef, aaf2.metadict.ClassDef)
            p = test_classdef.register_propertydef("TheAnswer", test_prop_auid, test_prop_pid, 'aafInt64', True, False)
            assert isinstance(p, aaf2.metadict.PropertyDef)

        with aaf2.open(result_file, 'r') as f:
            parent_classdef = f.metadict.lookup_classdef(parent_class_name)
            test_classdef = f.metadict.lookup_classdef('TestClass')

            self.assertTrue(test_classdef.auid == test_classdef_auid)
            self.assertTrue(test_classdef.uuid == UUID(bytes_le=bytes(test_classdef_auid.bytes_le)))
            self.assertTrue(test_classdef.class_name == "TestClass")
            self.assertTrue(test_classdef.parent is parent_classdef)

            p = test_classdef['Properties'].value[0]
            self.assertTrue(p.auid == test_prop_auid)
            self.assertTrue(p.pid == test_prop_pid)

            self.assertTrue(p.property_name == "TheAnswer")

            prop_typedef = f.metadict.lookup_typedef("aafInt64")
            self.assertTrue(p.typedef is prop_typedef)

    def test_add_extensions(self):
        result_file = common.get_test_file('basic_model.aaf')
        # create a file with basic model

        base_model_dynamic_pids = 0
        with aaf2.open(result_file, 'w', extensions=False) as f:
            for classdef in f.metadict['ClassDefinitions'].values():
                for pdef in classdef['Properties'].values():
                    if pdef['LocalIdentification'].value < 0x8000:
                        base_model_dynamic_pids += 1


        # check there are no extensions
        written_model_dynamic_pids = 0
        with aaf2.open(result_file, 'r') as f:
            for classdef in f.metadict['ClassDefinitions'].values():
                # make sure non of the properties and dynamics
                for pdef in classdef['Properties'].values():
                    if pdef['LocalIdentification'].value < 0x8000:
                        written_model_dynamic_pids +=1

        self.assertTrue(base_model_dynamic_pids == written_model_dynamic_pids)

        # open files 'rw' and add extensions
        with aaf2.open(result_file, 'rw', extensions=True) as f:
            pass

        dynamic_pids = set()
        # check that dynamic properties got added
        with aaf2.open(result_file, 'r') as f:
            for classdef in f.metadict['ClassDefinitions'].values():
                for pdef in classdef['Properties'].values():
                    if pdef['LocalIdentification'].value >= 0x8000:
                        pid = pdef['LocalIdentification'].value
                        self.assertTrue(pid not in dynamic_pids)
                        dynamic_pids.add(pid)

        assert len(dynamic_pids) > 0

    def test_duplicate_pids(self):
        result_file = common.get_test_file('duplicate_pids.aaf')
        with aaf2.open(result_file, 'w', extensions=True) as f:
            self.assertFalse(has_duplicate_pid(f))

    def test_duplicate_pids_add(self):
        result_file = common.get_test_file('duplicate_pids_add.aaf')
        with aaf2.open(result_file, 'w', extensions=False) as f:
            self.assertFalse(has_duplicate_pid(f))

        with aaf2.open(result_file, 'r') as f:
            self.assertFalse(has_duplicate_pid(f))

        with aaf2.open(result_file, 'rw', extensions=True) as f:
            self.assertFalse(has_duplicate_pid(f))

        with aaf2.open(result_file, 'r') as f:
            self.assertFalse(has_duplicate_pid(f))

    def test_duplicate_pids_existing(self):
        result_file = common.get_test_file('duplicate_pids_existing.aaf')
        shutil.copy(common.test_file_01(), result_file)

        with aaf2.open(result_file, 'w', extensions=False) as f:
            self.assertFalse(has_duplicate_pid(f))

        with aaf2.open(result_file, 'r') as f:
            self.assertFalse(has_duplicate_pid(f))

        with aaf2.open(result_file, 'rw', extensions=True) as f:
            self.assertFalse(has_duplicate_pid(f))

        with aaf2.open(result_file, 'r') as f:
            self.assertFalse(has_duplicate_pid(f))

    def test_add_extension_existing(self):
        test_file = common.get_test_file('extension_add_existing.aaf')
        shutil.copy(common.test_file_01(), test_file)

        original_pids = []
        original_classdefs_count = None
        original_typedef_count = None
        with aaf2.open(test_file, 'r') as f:
            original_classdefs_count = len(f.metadict['ClassDefinitions'])
            original_typedef_count = len(f.metadict['TypeDefinitions'])
            for classdef in f.metadict['ClassDefinitions'].values():
                # make sure non of the properties and dynamics
                for pdef in classdef['Properties'].values():
                    pid = pdef['LocalIdentification'].value
                    if pdef['LocalIdentification'].value >= 0x8000:
                        original_pids.append(pid)

        # open files 'rw' and add extensions
        with aaf2.open(test_file, 'rw', extensions=True) as f:
            pass

        all_pids = []
        with aaf2.open(test_file, 'r') as f:
            new_classdefs_count = len(f.metadict['ClassDefinitions'])
            new_typedef_count = len(f.metadict['TypeDefinitions'])
            for classdef in f.metadict['ClassDefinitions'].values():
                for pdef in classdef['Properties'].values():
                    pid = pdef['LocalIdentification'].value
                    # make sure there are no duplicate pids
                    self.assertTrue(pid not in all_pids)
                    all_pids.append(pid)

        self.assertTrue(new_classdefs_count > original_classdefs_count)
        self.assertTrue(new_typedef_count > original_typedef_count)

        # print(new_typedef_count, original_typedef_count)
        # print(new_classdefs_count, original_classdefs_count)


if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import unittest
import aaf2
from aaf2.auid import AUID
from aaf2 import audio
import common

def register_definitions(f):
    op_def = f.create.OperationDef('89d9b67e-5584-302d-9abd-8bd330c46841', 'VideoDissolve_2', '')
    f.dictionary.register_def(op_def)

    op_def.media_kind = 'picture'
    op_def['IsTimeWarp'].value = False
    op_def['Bypass'].value = 1
    op_def['NumberInputs'].value = 2
    op_def['OperationCategory'].value = 'OperationCategory_Effect'

    param_byteorder = f.create.ParameterDef("c0038672-a8cf-11d3-a05b-006094eb75cb", "AvidParameterByteOrder", "", 'aafUInt16')
    f.dictionary.register_def(param_byteorder)

    param_effect_id = f.create.ParameterDef("93994bd6-a81d-11d3-a05b-006094eb75cb", "AvidEffectID", "", 'AvidBagOfBits')
    f.dictionary.register_def(param_effect_id)

    op_def.parameters.extend([param_byteorder, param_effect_id])

    # note not part of VideoDissolve_2 op_def but still used...
    opacity_param = f.create.ParameterDef('8d56813d-847e-11d5-935a-50f857c10000', 'AFX_FG_KEY_OPACITY_U', '', 'Rational')
    f.dictionary.register_def(opacity_param)

    linear = f.create.InterpolationDef('5b6c85a4-0ede-11d3-80a9-006008143e6f', 'LinearInterp', '')
    f.dictionary.register_def(linear)

    op_def = f.create.OperationDef('0c3bea41-fc05-11d2-8a29-0050040ef7d2', 'Audio Dissolve', '')
    f.dictionary.register_def(op_def)

    op_def.media_kind = 'sound'
    op_def['IsTimeWarp'].value = False
    op_def['Bypass'].value = 1
    op_def['NumberInputs'].value = 2
    op_def['OperationCategory'].value = 'OperationCategory_Effect'
    op_def.parameters.extend([param_byteorder, param_effect_id])

def create_video_clip_import(f):
    frame_rate  = 25
    profile_name = 'dnx_1080p_36_25'
    sample = common.generate_dnxhd(profile_name, "transition.dnxhd", 200)

    tape = None
    master_mob = f.create.MasterMob("MasterMob")
    f.content.mobs.append(master_mob)
    slot = master_mob.import_dnxhd_essence(sample, frame_rate, tape=tape)

    return master_mob.create_source_clip(slot.slot_id, start=50)

def create_audio_clip_import(f):
    audio_profile_name = 'pcm_48000_s24le'
    frame_rate = 25
    frames = 200
    audio_duration = frames/frame_rate

    sample_format = audio.pcm_profiles[audio_profile_name]['sample_format']
    sample_rate = audio.pcm_profiles[audio_profile_name]['sample_rate']
    audio_sample = common.generate_pcm_audio_mono(audio_profile_name, sample_format=sample_format, sample_rate=sample_rate, duration=audio_duration)

    master_mob = f.create.MasterMob("MasterMob")
    f.content.mobs.append(master_mob)
    slot = master_mob.import_audio_essence(audio_sample, frame_rate)
    return master_mob.create_source_clip(slot.slot_id, start=50)

def create_video_clip_offline(f):

    master_mob = f.create.MasterMob("MasterMob")
    master_slot = master_mob.create_picture_slot()
    f.content.mobs.append(master_mob)

    source_mob = f.create.SourceMob()
    f.content.mobs.append(source_mob)

    descriptor = f.create.CDCIDescriptor()
    source_mob.descriptor = descriptor

    descriptor['ComponentWidth'].value = 8
    descriptor['HorizontalSubsampling'].value = 2
    descriptor['FrameLayout'].value = "FullFrame"
    descriptor['ImageAspectRatio'].value = "16/9"
    descriptor['StoredWidth'].value = 1920
    descriptor['StoredHeight'].value = 1080
    descriptor['Length'].value = 200

    descriptor['SampleRate'].value = 25
    descriptor['VideoLineMap'].value = [42, 0]
    descriptor['ContainerFormat'].value = f.dictionary.lookup_containerdef("AAF")
    dnxhd_codec_auid = aaf2.auid.AUID("8ef593f6-9521-4344-9ede-b84e8cfdc7da")
    descriptor['CodecDefinition'].value = f.dictionary.lookup_codecdef(dnxhd_codec_auid)

    source_slot = source_mob.create_picture_slot()
    clip = f.create.SourceClip()
    clip.length = 200
    source_slot.segment.components.append(clip)

    clip = source_mob.create_source_clip(slot_id= source_slot.slot_id, start = 0, length = 200)
    master_slot.segment.components.append(clip)

    return master_mob.create_source_clip(master_slot.slot_id, start=50)

def create_video_transition(f):

    length = 20
    cut_point = 10

    transition = f.create.Transition()
    transition.length = length
    transition['CutPoint'].value = cut_point

    op_group = f.create.OperationGroup('VideoDissolve_2')
    op_group.media_kind = 'picture'
    transition['OperationGroup'].value = op_group

    # now setup the dissovle curve
    opacity_u = f.create.VaryingValue("AFX_FG_KEY_OPACITY_U", "LinearInterp")

    # setting edit hint Proportional makes some avid warnings go away
    opacity_u.add_keyframe(0.0,   0.0, 'Proportional')
    opacity_u.add_keyframe(1.0, 100.0, 'Proportional')

    # not sure what the AUID is suppose to represent
    opacity_u['VVal_Extrapolation'].value = AUID('0e24dd54-66cd-4f1a-b0a0-670ac3a7a0b3')
    opacity_u['VVal_FieldCount'].value = 1
    op_group.parameters.append(opacity_u)

    op_group.length = length
    # transition.dump()
    return transition

def create_audio_transition(f):

    length = 40
    cut_point = 20

    transition = f.create.Transition('sound')
    transition.length = length
    transition['CutPoint'].value = cut_point

    op_group = f.create.OperationGroup('Audio Dissolve')
    op_group.media_kind = 'sound'
    transition['OperationGroup'].value = op_group

    return transition

class TestTransitions(unittest.TestCase):

    def test_video_transition(self):
        result_file = common.get_test_file('video_transition.aaf')

        with aaf2.open(result_file, 'w') as f:
            register_definitions(f)

            comp = f.create.CompositionMob("Transition Timeline")
            comp.usage = 'Usage_TopLevel'

            f.content.mobs.append(comp)

            slot = comp.create_picture_slot()
            seq = f.create.Sequence()
            slot.segment = seq

            filler = f.create.Filler()
            filler.length= 50
            seq.components.append(filler)

            transition = create_video_transition(f)
            seq.components.append(transition)

            clip = create_video_clip_import(f)
            clip.mob.name = "clip1"
            clip.length = 50
            seq.components.append(clip)

            transition = create_video_transition(f)
            seq.components.append(transition)

            clip = create_video_clip_offline(f)
            clip.length = 50
            clip.mob.name = "clip2"
            seq.components.append(clip)

            transition = create_video_transition(f)
            seq.components.append(transition)

            filler = f.create.Filler()
            filler.length= 50
            seq.components.append(filler)


        with aaf2.open(result_file, 'r') as f:
            comp = next(f.content.toplevel())

            slot = comp.slots[0]
            transitions = []
            for item in slot.segment.components:
                if isinstance(item, aaf2.components.Transition):
                    transitions.append(item)

            assert len(transitions) == 3
            for t in transitions:
                op_group = t['OperationGroup'].value
                opacity_u = None
                for p in op_group.parameters:
                    if p.name == 'AFX_FG_KEY_OPACITY_U':
                        opacity_u = p
                        break
                for point in opacity_u['PointList'].value:
                    assert isinstance(point['Value'].value, aaf2.rational.AAFRational)
            seq = slot.segment
            clip = seq.component_at_time(0)
            assert isinstance(clip, aaf2.components.Filler)
            clip = seq.component_at_time(29)
            assert isinstance(clip, aaf2.components.Filler)
            clip = seq.component_at_time(30)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(39)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(40)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(49)
            assert isinstance(clip, aaf2.components.Transition)

            clip = seq.component_at_time(50)
            assert isinstance(clip, aaf2.components.SourceClip)
            assert clip.mob.name == 'clip1'

            clip = seq.component_at_time(59)
            assert isinstance(clip, aaf2.components.SourceClip)
            assert clip.mob.name == 'clip1'

            clip = seq.component_at_time(60)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(69)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(70)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(79)
            assert isinstance(clip, aaf2.components.Transition)

            clip = seq.component_at_time(80)
            assert isinstance(clip, aaf2.components.SourceClip)
            assert clip.mob.name == 'clip2'

            clip = seq.component_at_time(89)
            assert isinstance(clip, aaf2.components.SourceClip)
            assert clip.mob.name == 'clip2'

            clip = seq.component_at_time(90)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(99)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(109)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(110)
            assert isinstance(clip, aaf2.components.Filler)
            clip = seq.component_at_time(139)
            assert isinstance(clip, aaf2.components.Filler)
            clip = seq.component_at_time(150)
            assert isinstance(clip, aaf2.components.Filler)

    def test_audio_transition(self):
        result_file = common.get_test_file('audio_transition.aaf')

        with aaf2.open(result_file, 'w') as f:
            register_definitions(f)

            comp = f.create.CompositionMob("Transition Timeline")
            comp.usage = 'Usage_TopLevel'

            f.content.mobs.append(comp)

            slot = comp.create_sound_slot()
            seq = f.create.Sequence('sound')
            slot.segment = seq

            filler = f.create.Filler('sound')
            filler.length = 50
            seq.components.append(filler)

            t = create_audio_transition(f)
            seq.components.append(t)

            clip = create_audio_clip_import(f)
            clip.mob.name = "clip1"
            clip.length = 100
            seq.components.append(clip)

            t = create_audio_transition(f)
            seq.components.append(t)

            filler = f.create.Filler('sound')
            filler.length = 50
            seq.components.append(filler)

        with aaf2.open(result_file, 'r') as f:
            comp = next(f.content.toplevel())

            slot = comp.slots[0]
            seq = slot.segment
            clip = seq.component_at_time(0)
            assert isinstance(clip, aaf2.components.Filler)
            clip = seq.component_at_time(9)
            assert isinstance(clip, aaf2.components.Filler)
            clip = seq.component_at_time(10)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(29)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(30)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(49)
            assert isinstance(clip, aaf2.components.Transition)

            clip = seq.component_at_time(50)
            assert isinstance(clip, aaf2.components.SourceClip)
            assert clip.mob.name == 'clip1'

            clip = seq.component_at_time(69)
            assert isinstance(clip, aaf2.components.SourceClip)
            assert clip.mob.name == 'clip1'

            clip = seq.component_at_time(70)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(89)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(90)
            assert isinstance(clip, aaf2.components.Transition)
            clip = seq.component_at_time(109)
            assert isinstance(clip, aaf2.components.Transition)

            clip = seq.component_at_time(110)
            assert isinstance(clip, aaf2.components.Filler)
            clip = seq.component_at_time(300)
            assert isinstance(clip, aaf2.components.Filler)





if __name__ == "__main__":
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from aaf2.cfb import CompoundFileBinary

from struct import unpack
import shutil
import os

base = os.path.join(os.path.dirname(os.path.abspath(__file__)))
test_dir = os.path.join(base, 'results')
if not os.path.exists(test_dir):
    os.makedirs(test_dir)

test_files = os.path.join(base, 'test_files')


if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.DEBUG)
    test_file_orig = os.path.join(test_files, "test_file_01.aaf")
    test_file = os.path.join(test_dir, "test_file_edit.aaf")

    shutil.copy(test_file_orig, test_file)

    f = open(test_file, 'rb+')

    ss = CompoundFileBinary(f)

    ss.rmtree("/MetaDictionary-1")

    ss.rmtree("/Header-2")
    ss.close()

    f = open(test_file, 'rb')
    ss = CompoundFileBinary(f)

    for root, storage, streams in ss.walk():
        print(root.path())
        for item in streams:
            # data = item.data()
            # print item
            print(item.path())



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import os
import subprocess
import unittest
import hashlib
import shutil
import common
import aaf2
from aaf2 import exceptions

import binascii

def stream_checksum(s):
    crc = 0
    f = s.open('r')
    while True:
        data = f.read(4092)
        if not data:
            break

        crc = binascii.crc32(data, crc) & 0xffffffff

    path = f.dir.path()

    return path, crc


class ImportTests(unittest.TestCase):

    def test_attach(self):
        result_file = common.get_test_file('attach.aaf')
        with aaf2.open(result_file, 'w') as f:

            mob = f.create.MasterMob()
            mob_id = mob.mob_id
            f.content.mobs.append(mob)

            with self.assertRaises(exceptions.AAFAttachError):
                f.content.mobs.append(mob)

            # dettach
            mob = f.content.mobs.pop(mob_id)
            assert mob
            assert len(f.content.mobs) == 0

            f.content.mobs.append(mob)

            slot = mob.create_timeline_slot(25)
            slot.segment = f.create.Sequence()

            comp_paths = []
            for i in range(10):
                filler = f.create.Filler()
                filler['DataDefinition'].value = f.dictionary.lookup_datadef("picture")
                filler['Length'].value = i
                slot.segment['Components'].append(filler)
                comp_paths.append(filler.dir.path())

            assert len(slot.segment['Components']) == 10

            last = slot.segment['Components'].pop(-1)
            assert len(slot.segment['Components']) == 9
            first = slot.segment['Components'].pop(0)
            assert len(slot.segment['Components']) == 8

            for filler in slot.segment['Components']:
                assert filler.dir

            with self.assertRaises(exceptions.AAFAttachError):
                slot.segment['Components'][0] = slot.segment['Components'][1]

            mob = f.content.mobs.pop(mob_id)


            for filler in slot.segment['Components']:
                assert filler.dir is None

            assert slot.dir == None
            assert len(slot.segment['Components']) == 8

            f.content.mobs.append(mob)

            slot.segment['Components'][0] = first

            assert slot.dir

    def test_reattach(self):
        new_file = os.path.join(common.sandbox(), 'test_reattach.aaf')
        test_file = common.test_file_01()
        shutil.copy(test_file, new_file)

        mob_ids = []
        original_child_count = 0

        checksums = {}

        with aaf2.open(new_file, 'r+') as f:
            mobs = f.content['Mobs'].value

            for m in mobs:
                for item, streams in m.walk_references():
                    assert item.dir is not None
                    original_child_count += 1

                    for item in streams:
                        path, crc = stream_checksum(item)
                        path = path.replace(m.dir.path(), str(m.mob_id))
                        checksums[path] = crc

            assert original_child_count > len(mobs)

            mob_ids = [m.mob_id for m in mobs]
            f.content['Mobs'].value = []

            child_count = 0
            for m in mobs:
                for item, streams in m.walk_references():
                    assert item.dir is None
                    child_count += 1

            assert child_count == original_child_count

            f.content['Mobs'].value = mobs
            child_count = 0
            for m in mobs:
                for item, streams in m.walk_references():
                    assert item.dir is not None
                    child_count += 1
                    for item in streams:
                        path, crc = stream_checksum(item)
                        path = path.replace(m.dir.path(), str(m.mob_id))
                        assert checksums[path] == crc

            assert child_count == original_child_count

        with aaf2.open(new_file, 'r') as f:
            for mob_id in mob_ids:
                assert f.content.mobs.get(mob_id, None) is not None

            mobs = f.content['Mobs'].value
            child_count = 0
            for m in mobs:
                for item, streams in m.walk_references():
                    assert item.dir is not None
                    child_count += 1

                    for item in streams:
                        path, crc = stream_checksum(item)
                        path = path.replace(m.dir.path(), str(m.mob_id))
                        assert checksums[path] == crc

            assert child_count == original_child_count

    def test_reattach512(self):
        new_file = os.path.join(common.sandbox(), 'test_reattach512.aaf')
        test_file = common.test_file_512()
        shutil.copy(test_file, new_file)

        mob_ids = []
        original_child_count = 0

        with aaf2.open(new_file, 'r+') as f:
            mobs = f.content['Mobs'].value

            for m in mobs:
                for item, streams in m.walk_references():
                    assert item.dir is not None
                    original_child_count += 1

            assert original_child_count > len(mobs)

            mob_ids = [m.mob_id for m in mobs]
            f.content['Mobs'].value = []

            child_count = 0
            for m in mobs:
                for item, streams in m.walk_references():
                    assert item.dir is None
                    child_count += 1

            assert child_count == original_child_count

            f.content['Mobs'].value = mobs
            child_count = 0
            for m in mobs:
                for item, streams in m.walk_references():
                    assert item.dir is not None
                    child_count += 1

            assert child_count == original_child_count

        with aaf2.open(new_file, 'r') as f:
            for mob_id in mob_ids:
                assert f.content.mobs.get(mob_id, None) is not None

            mobs = f.content['Mobs'].value
            child_count = 0
            for m in mobs:
                for item, streams in m.walk_references():
                    assert item.dir is not None
                    child_count += 1

            assert child_count == original_child_count

    def test_mob_id_swap(self):
        new_file = os.path.join(common.sandbox(), 'swap_id.aaf')
        test_file = common.test_file_01()
        shutil.copy(test_file, new_file)
        new_mobid =aaf2.mobid.MobID.new()
        with aaf2.open(new_file, 'rw') as f:
            comp = next(f.content.compositionmobs())
            comp.mob_id = new_mobid

        with aaf2.open(new_file, 'r') as f:
            comp = next(f.content.compositionmobs())
            assert comp.mob_id == new_mobid

    def test_rewrite(self):
        new_file = os.path.join(common.sandbox(), 'test_rewrite.aaf')
        test_file = common.test_file_01()
        shutil.copy(test_file, new_file)

        prop_count = {}

        with aaf2.open(new_file, 'rw') as f:
            for obj, streams in f.root.walk_references():
                f.manager.add_modified(obj)
                prop_count[obj.dir.path()] = len(list(obj.properties()))

        with aaf2.open(new_file, 'r') as f:
            for obj, streams in f.root.walk_references():
                path = obj.dir.path()
                assert len(list(obj.properties())) == prop_count[path]

    def test_rewrite512(self):
        new_file = os.path.join(common.sandbox(), 'test_rewrite512.aaf')
        test_file = common.test_file_512()
        shutil.copy(test_file, new_file)

        prop_count = {}

        with aaf2.open(new_file, 'rw') as f:
            for obj, streams in f.root.walk_references():
                f.manager.add_modified(obj)
                prop_count[obj.dir.path()] = len(list(obj.properties()))

        with aaf2.open(new_file, 'r') as f:
            for obj, streams in f.root.walk_references():
                path = obj.dir.path()
                assert len(list(obj.properties())) == prop_count[path]

if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from aaf2.mobid import MobID
from uuid import UUID
from aaf2.auid import AUID
import uuid
import unittest

class MobIDTests(unittest.TestCase):
    def test_mob_id(self):
        m = MobID.new()
        material_uuid = AUID("52c02cd8-6801-4806-986a-b68c0a0cf9d3")
        m.material = material_uuid
        m_str = "urn:smpte:umid:060a2b34.01010105.01010f20.13000000.52c02cd8.68014806.986ab68c.0a0cf9d3"

        m2 = MobID(str(m))

        assert m == m2
        m2 = MobID(bytes_le=m.bytes_le)
        assert m == m2
        assert m.int == m2.int

        assert m == MobID(m_str)
        assert hash(m) == hash(m2)
        assert str(m) == m_str

        assert m.material == material_uuid

    def test_int(self):

        for i in range(1000):
            m = MobID()
            m.int = i
            assert m.int == i

    def test_material_id(self):
        for i in range(10000):
            material = AUID(int=i)
            m = MobID(int=i)
            assert m.material == material

        for i in range(100):
            material = uuid.uuid4()
            m = MobID.new()
            m.material = material
            assert m.material == material




if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()
